import configparser
import json
import os
import random
from typing import TYPE_CHECKING, Any, Dict, List, Union

import pymysql
from itemadapter import ItemAdapter

from ayugespidertools.common.params import Param
from ayugespidertools.common.typevars import (
    AlterItem,
    AlterItemTable,
    MysqlConf,
    PostgreSQLConf,
)
from ayugespidertools.config import logger
from ayugespidertools.items import AyuItem

try:
    import psycopg
except ImportError:
    # pip install ayugespidertools[database]
    pass

__all__ = [
    "ReuseOperation",
]

if TYPE_CHECKING:
    from pathlib import Path

    from scrapy.settings import BaseSettings


class ReuseOperation:
    """用于存放经常复用的一些操作"""

    @staticmethod
    def fetch_local_conf(vit_dir: Union[str, "Path"], inner_settings: dict) -> dict:
        """通过本地 VIT 中的 .conf 获取所需配置，并将其添加到 inner_settings

        Args:
            vit_dir: 配置文件所在的目录
            inner_settings: inner_settings 配置

        Returns:
            inner_settings: 本库所需的配置
        """
        cfg = configparser.ConfigParser()
        cfg.read(f"{vit_dir}/.conf", encoding="utf-8")
        if "mysql" in cfg:
            mysql_section = cfg["mysql"]
            _charset = mysql_section.get("charset", "utf8mb4")
            inner_settings["MYSQL_CONFIG"] = {
                "host": mysql_section.get("host", "localhost"),
                "port": mysql_section.getint("port", 3306),
                "user": mysql_section.get("user", ""),
                "password": mysql_section.get("password", ""),
                "charset": _charset,
                "database": mysql_section.get("database", ""),
                "engine": mysql_section.get("engine", "InnoDB"),
                "collate": mysql_section.get(
                    "collate",
                    Param.charset_collate_map.get(_charset, "utf8mb4_general_ci"),
                ),
                "odku_enable": mysql_section.getboolean("odku_enable", False),
            }
        if "mongodb:uri" in cfg:
            inner_settings["MONGODB_CONFIG"] = {
                "uri": cfg.get("mongodb:uri", "uri", fallback=None)
            }
        elif "mongodb" in cfg:
            mongodb_section = cfg["mongodb"]
            inner_settings["MONGODB_CONFIG"] = {
                "host": mongodb_section.get("host", "localhost"),
                "port": mongodb_section.getint("port", 27017),
                "authsource": mongodb_section.get("authsource", "admin"),
                "authMechanism": mongodb_section.get("authMechanism", "SCRAM-SHA-1"),
                "user": mongodb_section.get("user", "admin"),
                "password": mongodb_section.get("password", None),
                "database": mongodb_section.get("database", None),
            }
        if "postgresql" in cfg:
            postgres_section = cfg["postgresql"]
            inner_settings["POSTGRESQL_CONFIG"] = {
                "host": postgres_section.get("host", "localhost"),
                "port": postgres_section.getint("port", 5432),
                "user": postgres_section.get("user", "postgres"),
                "password": postgres_section.get("password", ""),
                "database": postgres_section.get("database", ""),
                "charset": postgres_section.get("charset", "UTF8"),
            }
        if "elasticsearch" in cfg:
            es_section = cfg["elasticsearch"]
            inner_settings["ES_CONFIG"] = {
                "hosts": es_section.get("hosts", None),
                "index_class": json.loads(
                    es_section.get(
                        "index_class", '{"settings":{"number_of_shards": 2}}'
                    )
                ),
                "user": es_section.get("user", None),
                "password": es_section.get("password", None),
                "init": es_section.getboolean("init", False),
                "verify_certs": es_section.getboolean("verify_certs", False),
                "ca_certs": es_section.get("ca_certs", None),
                "client_cert": es_section.get("client_cert", None),
                "client_key": es_section.get("client_key", None),
                "ssl_assert_fingerprint": es_section.get(
                    "ssl_assert_fingerprint", None
                ),
            }
        if "oracle" in cfg:
            oracle_section = cfg["oracle"]
            inner_settings["ORACLE_CONFIG"] = {
                "host": oracle_section.get("host", "localhost"),
                "port": oracle_section.getint("port", 1521),
                "user": oracle_section.get("user", None),
                "password": oracle_section.get("password", None),
                "service_name": oracle_section.get("service_name", None),
                "encoding": oracle_section.get("encoding", "utf8"),
                "thick_lib_dir": oracle_section.get("thick_lib_dir", False),
            }
        if "consul" in cfg:
            consul_section = cfg["consul"]
            inner_settings["REMOTE_CONFIG"] = {
                "token": consul_section.get("token", None),
                "url": consul_section.get("url", None),
                "format": consul_section.get("format", "json"),
                "remote_type": "consul",
            }
        elif "nacos" in cfg:
            nacos_section = cfg["nacos"]
            inner_settings["REMOTE_CONFIG"] = {
                "token": nacos_section.get("token", None),
                "url": nacos_section.get("url", None),
                "format": nacos_section.get("format", "json"),
                "remote_type": "nacos",
            }
        if "kdl_dynamic_proxy" in cfg:
            kdl_dynamic_section = cfg["kdl_dynamic_proxy"]
            inner_settings["DYNAMIC_PROXY_CONFIG"] = {
                "proxy": kdl_dynamic_section.get("proxy", None),
                "username": kdl_dynamic_section.get("username", None),
                "password": kdl_dynamic_section.get("password", None),
            }
        if "kdl_exclusive_proxy" in cfg:
            kdl_exclusive_section = cfg["kdl_exclusive_proxy"]
            inner_settings["EXCLUSIVE_PROXY_CONFIG"] = {
                "proxy": kdl_exclusive_section.get("proxy", None),
                "username": kdl_exclusive_section.get("username", None),
                "password": kdl_exclusive_section.get("password", None),
                "index": kdl_exclusive_section.getint("index", 1),
            }
        if "mq" in cfg:
            mq_section = cfg["mq"]
            inner_settings["MQ_CONFIG"] = {
                "host": mq_section.get("host", "localhost"),
                "port": mq_section.getint("port", 5672),
                "username": mq_section.get("username", "guest"),
                "password": mq_section.get("password", "guest"),
                "virtualhost": mq_section.get("virtualhost", "/"),
                "heartbeat": mq_section.getint("heartbeat", 0),
                "socket_timeout": mq_section.getint("socket_timeout", 1),
                "queue": mq_section.get("queue", None),
                "durable": mq_section.getboolean("durable", True),
                "exclusive": mq_section.getboolean("exclusive", False),
                "auto_delete": mq_section.getboolean("auto_delete", False),
                "exchange": mq_section.get("exchange", None),
                "routing_key": mq_section.get("routing_key", None),
                "content_type": mq_section.getint("content_type", "text/plain"),
                "delivery_mode": mq_section.getint("delivery_mode", 1),
                "mandatory": mq_section.getboolean("mandatory", True),
            }
        if "kafka" in cfg:
            kafka_section = cfg["kafka"]
            inner_settings["KAFKA_CONFIG"] = {
                "bootstrap_servers": kafka_section.get(
                    "bootstrap_servers", "127.0.0.1:9092"
                ),
                "topic": kafka_section.get("topic", None),
                "key": kafka_section.get("key", None),
            }
        if "oss:ali" in cfg:
            oss_section = cfg["oss:ali"]
            inner_settings["OSS_CONFIG"] = {
                "access_key": oss_section.get("access_key", None),
                "access_secret": oss_section.get("access_secret", None),
                "endpoint": oss_section.get("endpoint", None),
                "bucket": oss_section.get("bucket", None),
                "doc": oss_section.get("doc", None),
                "upload_fields_suffix": oss_section.get(
                    "upload_fields_suffix", "_file_url"
                ),
                "oss_fields_prefix": oss_section.get("oss_fields_prefix", "_"),
            }
        return inner_settings

    @staticmethod
    def item_to_dict(item: Union[AyuItem, dict]) -> dict:
        """将 item 转换为 dict 类型；
        将 spider 中的 yield 的 item 转换为 dict 类型，方便后续处理

        Args:
            item: spider 中的 yield 的 item

        Returns:
            1). dict 类型的 item
        """
        return (
            item.asdict() if isinstance(item, AyuItem) else ItemAdapter(item).asdict()
        )

    @classmethod
    def reshape_item(cls, item_dict: Dict[str, Any]) -> AlterItem:
        """重新整合 item

        Args:
            item_dict: dict 类型的 item

        Returns:
            1). 整合后的 item
        """
        new_item = {}
        notes_dic = {}

        insert_data = cls.get_items_except_keys(
            dict_conf=item_dict, keys=["_mongo_update_rule", "_table"]
        )
        judge_item = next(iter(insert_data.values()))
        # 是 namedtuple 类型
        if cls.is_namedtuple_instance(judge_item):
            _table_name = item_dict["_table"].key_value
            _table_notes = item_dict["_table"].notes
            table_info = AlterItemTable(_table_name, _table_notes)
            for key, value in insert_data.items():
                new_item[key] = value.key_value
                notes_dic[key] = value.notes
        # 是普通的 dict 类型
        else:
            _table_name = item_dict["_table"]
            table_info = AlterItemTable(_table_name, "")
            for key, value in insert_data.items():
                new_item[key] = value
                notes_dic[key] = ""

        return AlterItem(new_item, notes_dic, table_info)

    @staticmethod
    def is_namedtuple_instance(x: Any) -> bool:
        """判断 x 是否为 namedtuple 类型

        Args:
            x: 需要判断的参数

        Returns:
            1). 是否符合 namedtuple 类型
        """
        return isinstance(x, tuple) and hasattr(x, "_fields")

    @staticmethod
    def get_files_from_path(path: str) -> list:
        """获取 path 文件夹下的所有文件，并输出以 path 为根目录的相对路径

        Args:
            path: 需要判断的文件夹路径

        Returns:
            1). path 文件夹下的文件列表
        """
        return [f.path for f in os.scandir(path) if f.is_file()]

    @staticmethod
    def random_weight(weight_data: list):
        """带权重的随机取值，即在带权重的列表数据中根据权重随机取一个值

        Args:
            weight_data: 带权重的列表信息，示例：
                [{'username': 'xxxx', 'password': '******', 'weight': 8}, ...]

        Returns:
            ret: 返回当前权重列表 account_arr 中的一个值
        """
        total = sum(item["weight"] for item in weight_data)
        # 在 0 与权重和之间获取一个随机数
        ra = random.uniform(0, total)
        curr_sum = 0
        ret = None
        for data in weight_data:
            # 在遍历中，累加当前权重值
            curr_sum += data["weight"]
            # 当随机数 <= 当前权重和时，返回权重 key
            if ra <= curr_sum:
                ret = data
                break
        return ret

    @classmethod
    def is_dict_meet_min_limit(cls, dict_conf: dict, key_list: List[str]) -> bool:
        """判断 dict_conf 是否满足 key_list 中的 key 值限定

        Args:
            dict_conf: 需要判断的参数
            key_list: dict_conf 中需要包含的 key 值列表，示例为：['proxy', 'username', 'password']

        Returns:
            1). 是否满足 key 值限制
        """
        if any([not dict_conf, not isinstance(dict_conf, dict)]):
            return False

        return all(key in dict_conf for key in key_list)

    @classmethod
    def get_items_by_keys(
        cls,
        dict_conf: dict,
        keys: List[str],
    ) -> dict:
        """获取 dict_conf 中的含有 keys 的 key 的字段

        Args:
            dict_conf: 需要处理的参数
            keys: 需要取的 key 值列表

        Returns:
            1). 取值后的 dict，或不满足请求的 False 值
        """
        # 参数先要满足最小限定，然后再取出限定的参数值；否则返回空字典
        return (
            {k: dict_conf[k] for k in keys}
            if cls.is_dict_meet_min_limit(dict_conf=dict_conf, key_list=keys)
            else {}
        )

    @classmethod
    def get_items_except_keys(cls, dict_conf, keys: List[str]) -> dict:
        """获取 dict_conf 中的不含有 keys 的 key 的字段

        Args:
            dict_conf: 需要处理的参数
            keys: 需要排除的 key 值列表

        Returns:
            1). dict_conf 排除 keys 中的键值后的值
        """
        return {k: dict_conf[k] for k in dict_conf if k not in keys}

    @classmethod
    def create_database(cls, db_conf: Union[MysqlConf, PostgreSQLConf]) -> None:
        """创建数据库：由于这是在连接数据库，报数据库不存在错误时的场景，则需要
        新建(不指定数据库)连接创建好所需数据库即可

        Args:
            db_conf: 数据库连接配置，目前支持 mysql 和 postgresql
        """
        if isinstance(db_conf, MysqlConf):
            with pymysql.connect(
                user=db_conf.user,
                password=db_conf.password,
                host=db_conf.host,
                port=db_conf.port,
                charset=db_conf.charset,
            ) as conn:
                with conn.cursor() as cur:
                    cur.execute(
                        f"CREATE DATABASE IF NOT EXISTS `{db_conf.database}` character set {db_conf.charset};"
                    )

        elif isinstance(db_conf, PostgreSQLConf):
            with psycopg.connect(
                user=db_conf.user,
                password=db_conf.password,
                host=db_conf.host,
                port=db_conf.port,
            ) as conn:
                with conn.cursor() as cur:
                    conn.autocommit = True
                    cur.execute(
                        f"CREATE DATABASE {db_conf.database} WITH ENCODING {db_conf.charset};"
                    )

        else:
            assert False, f"Invalid db_conf type: {type(db_conf)}"
        logger.info(f"创建数据库 {db_conf.database} 成功，其 charset 类型是：{db_conf.charset}!")

    @classmethod
    def dict_keys_to_lower(cls, deal_dict: dict) -> dict:
        """将 dict 中 str 类型的 key 值变成小写

        Args:
            deal_dict: 需要处理的 dict

        Returns:
            1).处理后的 dict 值
        """
        key_to_lower_dict = {}
        for key, value in deal_dict.items():
            if isinstance(value, dict):
                if isinstance(key, str):
                    key_to_lower_dict[key.lower()] = cls.dict_keys_to_lower(value)
                else:
                    key_to_lower_dict[key] = cls.dict_keys_to_lower(value)
            elif isinstance(key, str):
                key_to_lower_dict[key.lower()] = value
            else:
                key_to_lower_dict[key] = value
        return key_to_lower_dict

    @classmethod
    def dict_keys_to_upper(cls, deal_dict: dict) -> dict:
        """将 dict 中 str 类型的 key 值变成大写

        Args:
            deal_dict: 需要处理的 dict

        Returns:
            1).处理后的 dict 值
        """
        key_to_upper_dict = {}
        for key, value in deal_dict.items():
            if isinstance(value, dict):
                if isinstance(key, str):
                    key_to_upper_dict[key.upper()] = cls.dict_keys_to_upper(value)
                else:
                    key_to_upper_dict[key] = cls.dict_keys_to_upper(value)
            elif isinstance(key, str):
                key_to_upper_dict[key.upper()] = value
            else:
                key_to_upper_dict[key] = value
        return key_to_upper_dict

    @classmethod
    def get_remote_option(cls, settings: "BaseSettings") -> dict:
        """获取项目中 consul 或 nacos 的链接配置

        Args:
            settings: scrapy 的 settings 信息

        Returns:
            1). 满足最少要求的远程配置
        """
        consul_conf_dict = settings.get("REMOTE_CONFIG", {})
        return cls.get_items_by_keys(
            dict_conf=consul_conf_dict, keys=["token", "url", "format", "remote_type"]
        )

    @classmethod
    def judge_str_is_json(cls, judge_str: str) -> bool:
        """判断字符串是否为 json 格式

        Args:
            judge_str: 需要判断的字符串

        Returns:
            1). 是否为 json 格式
        """
        if not isinstance(judge_str, str):
            return False

        try:
            json.loads(judge_str)
        except Exception:
            return False
        else:
            return True

    @staticmethod
    def get_ck_dict_from_headers(headers_ck_str: str) -> dict:
        """从 headers 中的 ck str 格式转化为 dict 格式

        Args:
            headers_ck_str: request headers ck 的 str 格式

        Returns:
            1). 转化 dict 格式后的 ck
        """
        # 也可以这样写，但不推荐
        # dict(line.split("=", 1) for line in headers_ck_str.split("; "))
        return {
            x.split("=", 1)[0].strip(): x.split("=", 1)[1].strip()
            for x in headers_ck_str.split(";")
        }

    @staticmethod
    def get_req_dict_from_scrapy(req_body_data_str: str) -> dict:
        """将 scrapy 请求中的 body 对象转为 dict 格式

        Args:
            req_body_data_str: scrapy 中的 body 参数

        Returns:
            1). 转化 dict 格式后的 body
        """
        return {
            x.split("=", 1)[0]: x.split("=", 1)[1] for x in req_body_data_str.split("&")
        }

    @classmethod
    def get_array_depth(cls, array: list) -> int:
        """获取 array 的最大层级，深度

        Args:
            array: 数组

        Returns:
            1). 最大层级，深度
        """

        """1 + max(map(depthCount,x)) if x and isinstance(x,list) else 0"""
        # 先判断是否为数组类型的元素
        judge_array = isinstance(
            array,
            (
                frozenset,
                list,
                set,
                tuple,
            ),
        )
        return (
            int(judge_array) and len(array) and 1 + max(map(cls.get_array_depth, array))
        )



# Define here the models for your spider middleware
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/spider-middleware.html
from ayugespidertools.scraper.middlewares.headers.ua import RandomRequestUaMiddleware
from ayugespidertools.scraper.middlewares.netlib.aiohttplib import (
    AiohttpDownloaderMiddleware,
)
from ayugespidertools.scraper.middlewares.proxy.dynamic import (
    AbuDynamicProxyDownloaderMiddleware,
    DynamicProxyDownloaderMiddleware,
)
from ayugespidertools.scraper.middlewares.proxy.exclusive import (
    ExclusiveProxyDownloaderMiddleware,
)

__all__ = [
    "RandomRequestUaMiddleware",
    "AiohttpDownloaderMiddleware",
    "DynamicProxyDownloaderMiddleware",
    "AbuDynamicProxyDownloaderMiddleware",
    "ExclusiveProxyDownloaderMiddleware",
]



from abc import ABCMeta
from collections.abc import MutableMapping
from dataclasses import dataclass
from typing import Any, Dict, NamedTuple, Optional, Union

import scrapy
from scrapy.item import Item

from ayugespidertools.common.typevars import EmptyKeyError, FieldAlreadyExistsError

__all__ = [
    "DataItem",
    "AyuItem",
]


class ScrapyItem(Item):
    """scrapy item 的标准方式"""

    ...


class DataItem(NamedTuple):
    """用于描述 item 中字段

    Attributes:
        key_value: 参数值
        notes: 对参数的注释
    """

    key_value: Any
    notes: Any = ""


class ItemMeta(ABCMeta):
    def __new__(cls, class_name, bases, attrs):
        def add_field(self, key: str, value: Any) -> None:
            """动态添加字段方法

            Args:
                self: self
                key: 需要添加的字段名
                value: 需要添加的字段对应的值
            """
            if not key:
                raise EmptyKeyError()
            if key in self._AyuItem__fields:
                raise FieldAlreadyExistsError(key)
            setattr(self, key, value)
            self._AyuItem__fields.add(key)

        def asdict(self) -> Dict[str, Any]:
            """将 AyuItem 转换为 dict"""
            self._AyuItem__fields.discard("_AyuItem__fields")
            _item_dict = {key: getattr(self, key) for key in self._AyuItem__fields}
            return _item_dict

        def asitem(self, assignment: bool = True) -> ScrapyItem:
            """将 AyuItem 转换为 ScrapyItem

            Args:
                self: self
                assignment: 是否将 AyuItem 中的值赋值给 ScrapyItem，默认为 True

            Returns:
                new_class: 转换 ScrapyItem 后的实例
            """
            item_temp = ScrapyItem()
            for k, v in self.asdict().items():
                item_temp.fields[k] = scrapy.Field()
                if assignment:
                    item_temp[k] = v
            return item_temp

        attrs["add_field"] = add_field
        attrs["asdict"] = asdict
        attrs["asitem"] = asitem
        return super().__new__(cls, class_name, bases, attrs)


@dataclass
class AyuItem(MutableMapping, metaclass=ItemMeta):
    """Used to create AyuItem and add fields dynamically,
    and provides methods to convert to dict and ScrapyItem.

    Examples:
        >>> item = AyuItem(
        ...     _table="ta",
        ... )
        >>> # 获取字段；
        >>> item["_table"]
        'ta'
        >>>
        >>> # 添加 / 修改字段，不存在则创建，存在则修改：
        >>> item["_table"] = "tab"
        >>> item["title"] = "tit"
        >>> # 也可通过 add_field 添加字段，但不能重复添加相同字段
        >>> item.add_field("num", 10)
        >>> [ item["_table"], item["title"], item["num"] ]
        ['tab', 'tit', 10]
        >>> item.asdict()
        {'title': 'tit', '_table': 'tab', 'num': 10}
        >>> type(item.asitem())
        <class 'ayugespidertools.items.ScrapyItem'>
        >>> # 删除字段：
        >>> del item["title"]
        >>> item
        {'_table': 'tab', 'num': 10}
    """

    def __init__(
        self,
        _table: Union[DataItem, str],
        _mongo_update_rule: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """初始化 AyuItem 实例

        Args:
            _table: 数据库表名。
            _mongo_update_rule: MongoDB 存储场景下可能需要的查重条件，默认为 None。
        """
        self.__fields = set()
        if _table:
            self.__fields.add("_table")
            setattr(self, "_table", _table)
        if _mongo_update_rule:
            self.__fields.add("_mongo_update_rule")
            setattr(self, "_mongo_update_rule", _mongo_update_rule)
        for key, value in kwargs.items():
            setattr(self, key, value)
            self.__fields.add(key)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        if key not in self.__fields:
            setattr(self, key, value)
            self.__fields.add(key)
        else:
            setattr(self, key, value)

    def __delitem__(self, key):
        if key not in self.__fields:
            raise KeyError(f"{key} not found")
        delattr(self, key)
        self.__fields.discard(key)

    def __getattr__(self, name):
        if name in self.__fields:
            raise AttributeError(f"Try to use item[{name!r}] to get field value")
        raise AttributeError(name)

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        self.__fields.add(name)

    def __delattr__(self, name):
        super().__delattr__(name)
        self.__fields.discard(name)

    def __iter__(self):
        return iter(self.__fields)

    def __len__(self):
        return len(self.__fields)

    def __str__(self):
        # 与下方 __repr__ 一样，不返回 AyuItem(field=data) 的格式
        return f"{self.asdict()}"

    def __repr__(self):
        return f"{self.asdict()}"

    def fields(self):
        self.__fields.discard("_AyuItem__fields")
        return self.__fields

    def add_field(self, key: str, value: Any) -> None:
        ...

    def asdict(self) -> Dict[str, Any]:
        ...

    def asitem(self, assignment: bool = True) -> ScrapyItem:
        ...



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, Optional, Tuple, TypeVar

from sqlalchemy.exc import OperationalError

from ayugespidertools.common.typevars import (
    DatabaseEngineClass,
    DynamicProxyConf,
    ESConf,
    ExclusiveProxyConf,
    KafkaConf,
    MongoDBConf,
    MQConf,
    MysqlConf,
    OracleConf,
    OssConf,
    PostgreSQLConf,
)
from ayugespidertools.common.utils import ToolsForAyu
from ayugespidertools.config import logger

try:
    from elasticsearch_dsl import connections
    from oracledb.exceptions import DatabaseError
except ImportError:
    # pip install ayugespidertools[database]  # oracledb or elasticsearch_dsl ImportError
    pass

__all__ = [
    "get_spider_conf",
    "ESConfCreator",
    "MysqlConfCreator",
    "MongoDBConfCreator",
    "MQConfCreator",
    "OracleConfCreator",
    "PostgreSQLConfCreator",
    "KafkaConfCreator",
    "DynamicProxyCreator",
    "ExclusiveProxyCreator",
    "OssConfCreator",
    "get_sqlalchemy_conf",
]

if TYPE_CHECKING:
    from elasticsearch import Elasticsearch
    from scrapy.settings import Settings
    from sqlalchemy.engine.base import Connection as SqlalchemyConnectT
    from sqlalchemy.engine.base import Engine as SqlalchemyEngineT

SpiderConf = TypeVar(
    "SpiderConf",
    MysqlConf,
    MongoDBConf,
    MQConf,
    KafkaConf,
    OssConf,
    DynamicProxyConf,
    ExclusiveProxyConf,
)


class Product(ABC, Generic[SpiderConf]):
    @abstractmethod
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[SpiderConf]:
        """获取各个工具链接配置信息"""
        pass

    @abstractmethod
    def get_engine(
        self, db_conf: SpiderConf, db_engine_enabled: bool
    ) -> Tuple[Optional["SqlalchemyEngineT"], Optional["SqlalchemyConnectT"]]:
        """获取各个工具中对应的 sqlalchemy db_engine 和 db_engine_conn。
        需要此方法的工具有 mysql，postgresql，oracle，elasticsearch 其余的不需要。
        但其中 elasticsearch 不采用 sqlalchemy 的方式。
        """
        pass


class Creator(ABC):
    @abstractmethod
    def create_product(self) -> Product:
        pass


class MysqlConfProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[MysqlConf]:
        # 1). 优先从远程配置中取值
        if settings.get("APP_CONF_MANAGE", False):
            remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="mysql", **remote_option
            )
            return MysqlConf(**remote_conf) if remote_conf else None

        # 2). 从本地参数中取值
        local_conf = settings.get("MYSQL_CONFIG")
        return MysqlConf(**local_conf) if local_conf else None

    def get_engine(
        self, db_conf: MysqlConf, db_engine_enabled: bool
    ) -> Tuple[Optional["SqlalchemyEngineT"], Optional["SqlalchemyConnectT"]]:
        mysql_engine = mysql_engine_conn = None
        if db_engine_enabled:
            mysql_url = (
                f"mysql+pymysql://{db_conf.user}"
                f":{db_conf.password}@{db_conf.host}"
                f":{db_conf.port}/{db_conf.database}"
                f"?charset={db_conf.charset}"
            )
            try:
                mysql_engine = DatabaseEngineClass(engine_url=mysql_url).engine
                mysql_engine_conn = mysql_engine.connect()
            except OperationalError as err:
                logger.warning(f"MySQL engine enabled failed: {err}")
        return mysql_engine, mysql_engine_conn


class MongoDBConfProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[MongoDBConf]:
        if settings.get("APP_CONF_MANAGE", False):
            uri_remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="mongodb:uri", **remote_option
            )
            normal_remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="mongodb", **remote_option
            )
            remote_conf = uri_remote_conf or normal_remote_conf
            return MongoDBConf(**remote_conf) if remote_conf else None

        local_conf = settings.get("MONGODB_CONFIG")
        return MongoDBConf(**local_conf) if local_conf else None

    def get_engine(self, db_conf: MongoDBConf, db_engine_enabled: bool):
        pass


class PostgreSQLConfProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[PostgreSQLConf]:
        if settings.get("APP_CONF_MANAGE", False):
            remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="postgresql", **remote_option
            )
            return PostgreSQLConf(**remote_conf) if remote_conf else None

        local_conf = settings.get("POSTGRESQL_CONFIG")
        return PostgreSQLConf(**local_conf) if local_conf else None

    def get_engine(
        self, db_conf: PostgreSQLConf, db_engine_enabled: bool
    ) -> Tuple[Optional["SqlalchemyEngineT"], Optional["SqlalchemyConnectT"]]:
        postgres_engine = postgres_engine_conn = None
        if db_engine_enabled:
            postgres_url = (
                f"postgresql+psycopg://{db_conf.user}:{db_conf.password}"
                f"@{db_conf.host}:{db_conf.port}/{db_conf.database}"
            )
            try:
                postgres_engine = DatabaseEngineClass(engine_url=postgres_url).engine
                postgres_engine_conn = postgres_engine.connect()
            except OperationalError as err:
                logger.warning(f"PostgreSQL engine enabled failed: {err}")
        return postgres_engine, postgres_engine_conn


class ESConfProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[ESConf]:
        if settings.get("APP_CONF_MANAGE", False):
            remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="elasticsearch", **remote_option
            )
            return ESConf(**remote_conf) if remote_conf else None

        local_conf = settings.get("ES_CONFIG")
        return ESConf(**local_conf) if local_conf else None

    def get_engine(  # type: ignore[override]
        self, db_conf: ESConf, db_engine_enabled: bool
    ) -> Optional["Elasticsearch"]:  # @override to fix mypy [override] error.
        if db_engine_enabled:
            _hosts_lst = db_conf.hosts.split(",")
            if any([db_conf.user is not None, db_conf.password is not None]):
                http_auth = (db_conf.user, db_conf.password)
            else:
                http_auth = None
            return connections.create_connection(
                hosts=_hosts_lst,
                http_auth=http_auth,
                verify_certs=db_conf.verify_certs,
                ca_certs=db_conf.ca_certs,
                client_cert=db_conf.client_cert,
                client_key=db_conf.client_key,
                ssl_assert_fingerprint=db_conf.ssl_assert_fingerprint,
            )
        return None


class OracleConfProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[OracleConf]:
        if settings.get("APP_CONF_MANAGE", False):
            remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="oracle", **remote_option
            )
            return OracleConf(**remote_conf) if remote_conf else None

        local_conf = settings.get("ORACLE_CONFIG")
        return OracleConf(**local_conf) if local_conf else None

    def get_engine(
        self, db_conf: OracleConf, db_engine_enabled: bool
    ) -> Tuple[Optional["SqlalchemyEngineT"], Optional["SqlalchemyConnectT"]]:
        oracle_engine = oracle_engine_conn = None
        if db_engine_enabled:
            oracle_url = (
                f"oracle+oracledb://{db_conf.user}:{db_conf.password}"
                f"@{db_conf.host}:{db_conf.port}/{db_conf.service_name}"
            )
            thick_mode = (
                {"lib_dir": db_conf.thick_lib_dir} if db_conf.thick_lib_dir else False
            )

            try:
                oracle_engine = DatabaseEngineClass(
                    engine_url=oracle_url,
                    thick_mode=thick_mode,
                ).engine
                oracle_engine_conn = oracle_engine.connect()
            except (OperationalError, DatabaseError) as err:
                logger.warning(f"Oracle engine enabled failed: {err}")
        return oracle_engine, oracle_engine_conn


class MQConfProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[MQConf]:
        if settings.get("APP_CONF_MANAGE", False):
            remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="rabbitmq", **remote_option
            )
            return MQConf(**remote_conf) if remote_conf else None

        local_conf = settings.get("MQ_CONFIG")
        return MQConf(**local_conf) if local_conf else None

    def get_engine(self, db_conf: MQConf, db_engine_enabled: bool):
        pass


class KafkaConfProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[KafkaConf]:
        if settings.get("APP_CONF_MANAGE", False):
            remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="kafka", **remote_option
            )
            return KafkaConf(**remote_conf) if remote_conf else None

        local_conf = settings.get("KAFKA_CONFIG")
        return KafkaConf(**local_conf) if local_conf else None

    def get_engine(self, db_conf: KafkaConf, db_engine_enabled: bool):
        pass


class DynamicProxyProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[DynamicProxyConf]:
        if settings.get("APP_CONF_MANAGE", False):
            remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="dynamicproxy", **remote_option
            )
            return DynamicProxyConf(**remote_conf) if remote_conf else None

        local_conf = settings.get("DYNAMIC_PROXY_CONFIG")
        return DynamicProxyConf(**local_conf) if local_conf else None

    def get_engine(self, db_conf: DynamicProxyConf, db_engine_enabled: bool):
        pass


class ExclusiveProxyProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[ExclusiveProxyConf]:
        if settings.get("APP_CONF_MANAGE", False):
            remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="exclusiveproxy", **remote_option
            )
            return ExclusiveProxyConf(**remote_conf) if remote_conf else None

        local_conf = settings.get("EXCLUSIVE_PROXY_CONFIG")
        return ExclusiveProxyConf(**local_conf) if local_conf else None

    def get_engine(self, db_conf: ExclusiveProxyConf, db_engine_enabled: bool):
        pass


class OssConfProduct(Product):
    def get_conn_conf(
        self, settings: "Settings", remote_option: dict
    ) -> Optional[OssConf]:
        if settings.get("APP_CONF_MANAGE", False):
            remote_conf = ToolsForAyu.fetch_remote_conf(
                conf_name="oss:ali", **remote_option
            )
            return OssConf(**remote_conf) if remote_conf else None

        local_conf = settings.get("OSS_CONFIG")
        return OssConf(**local_conf) if local_conf else None

    def get_engine(self, db_conf: OssConf, db_engine_enabled: bool):
        pass


class MysqlConfCreator(Creator):
    def create_product(self) -> Product:
        return MysqlConfProduct()


class MongoDBConfCreator(Creator):
    def create_product(self) -> Product:
        return MongoDBConfProduct()


class PostgreSQLConfCreator(Creator):
    def create_product(self) -> Product:
        return PostgreSQLConfProduct()


class ESConfCreator(Creator):
    def create_product(self) -> Product:
        return ESConfProduct()


class OracleConfCreator(Creator):
    def create_product(self) -> Product:
        return OracleConfProduct()


class MQConfCreator(Creator):
    def create_product(self) -> Product:
        return MQConfProduct()


class KafkaConfCreator(Creator):
    def create_product(self) -> Product:
        return KafkaConfProduct()


class DynamicProxyCreator(Creator):
    def create_product(self) -> Product:
        return DynamicProxyProduct()


class ExclusiveProxyCreator(Creator):
    def create_product(self) -> Product:
        return ExclusiveProxyProduct()


class OssConfCreator(Creator):
    def create_product(self) -> Product:
        return OssConfProduct()


def get_spider_conf(
    creator: Creator, settings: "Settings", remote_option: dict
) -> Optional[SpiderConf]:
    product = creator.create_product()
    return product.get_conn_conf(settings, remote_option)


def get_sqlalchemy_conf(
    creator: Creator,
    db_conf: SpiderConf,
    db_engine_enabled: bool,
) -> Tuple[Optional["SqlalchemyEngineT"], Optional["SqlalchemyConnectT"]]:
    product = creator.create_product()
    return product.get_engine(db_conf, db_engine_enabled)



from typing import TYPE_CHECKING

from pymysql import cursors
from twisted.enterprise import adbapi

from ayugespidertools.common.expend import MysqlPipeEnhanceMixin
from ayugespidertools.common.multiplexing import ReuseOperation
from ayugespidertools.common.mysqlerrhandle import TwistedAsynchronous, deal_mysql_err

__all__ = [
    "AyuTwistedMysqlPipeline",
]

if TYPE_CHECKING:
    from ayugespidertools.common.typevars import MysqlConf, slogT


class AyuTwistedMysqlPipeline(MysqlPipeEnhanceMixin):
    mysql_conf: "MysqlConf"
    slog: "slogT"
    dbpool: "adbapi.ConnectionPool"

    def open_spider(self, spider):
        assert hasattr(spider, "mysql_conf"), "未配置 Mysql 连接信息！"
        self.slog = spider.slog
        self.mysql_conf = spider.mysql_conf
        # 判断目标数据库是否连接正常。若连接目标数据库错误时，创建缺失的目标数据库。
        self._connect(self.mysql_conf).close()

        _mysql_conf = {
            "user": self.mysql_conf.user,
            "password": self.mysql_conf.password,
            "host": self.mysql_conf.host,
            "port": self.mysql_conf.port,
            "db": self.mysql_conf.database,
            "charset": self.mysql_conf.charset,
            "cursorclass": cursors.DictCursor,
        }
        self.dbpool = adbapi.ConnectionPool("pymysql", cp_reconnect=True, **_mysql_conf)
        query = self.dbpool.runInteraction(self.db_create)
        query.addErrback(self.db_create_err)

    def db_create(self, cursor):
        pass

    def db_create_err(self, failure):
        self.slog.error(f"创建数据表失败: {failure}")

    def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        query = self.dbpool.runInteraction(self.db_insert, item_dict)
        query.addErrback(self.handle_error, item)
        return item

    def db_insert(self, cursor, item):
        alter_item = ReuseOperation.reshape_item(item)
        if not (new_item := alter_item.new_item):
            return

        _table_name = alter_item.table.name
        _table_notes = alter_item.table.notes
        note_dic = alter_item.notes_dic
        sql, args = self._get_sql_by_item(
            table=_table_name,
            item=new_item,
            odku_enable=self.mysql_conf.odku_enable,
        )

        try:
            cursor.execute(sql, args)
        except Exception as e:
            self.slog.warning(
                f"Pipe Warn: {e} & Table: {_table_name} & Item: {new_item}"
            )
            deal_mysql_err(
                TwistedAsynchronous(),
                err_msg=str(e),
                cursor=cursor,
                mysql_conf=self.mysql_conf,
                table=_table_name,
                table_notes=_table_notes,
                note_dic=note_dic,
            )
            return self.db_insert(cursor, item)
        return item

    def handle_error(self, failure, item):
        self.slog.error(f"插入数据失败:{failure}, item: {item}")



from pathlib import Path

from loguru import logger

__all__ = [
    "NormalConfig",
    "logger",
]


class NormalConfig:
    """用于存放此项目的通用配置"""

    # 项目根目录及其它所需目录
    CONFIG_DIR = Path(__file__).parent
    ROOT_DIR = CONFIG_DIR.parent
    COMMON_DIR = CONFIG_DIR / "common"
    VIT_DIR = CONFIG_DIR / "VIT"



import random

from scrapy import signals

from ayugespidertools.common.params import Param

__all__ = [
    "RandomRequestUaMiddleware",
]


class RandomRequestUaMiddleware:
    """随机请求头中间件"""

    def __init__(self):
        self.explorer_types = None
        self.explorer_weights = None

    def get_random_ua_by_weight(self) -> str:
        # 先按权重取出所需浏览器类型
        explorer_types = random.choices(
            self.explorer_types, weights=self.explorer_weights
        )
        return random.choice(Param.fake_useragent_dict[explorer_types[0]])

    @classmethod
    def from_crawler(cls, crawler):
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def spider_opened(self, spider):
        # 带权重的 ua 列表，这里是根据 fake_useragent 库中的打印信息来规划权重的。
        ua_arr = [
            {"explorer": "opera", "weight": 16},
            {"explorer": "safari", "weight": 32},
            {"explorer": "internetexplorer", "weight": 41},
            {"explorer": "firefox", "weight": 124},
            {"explorer": "chrome", "weight": 772},
        ]
        self.explorer_types = [x["explorer"] for x in ua_arr]
        self.explorer_weights = [x["weight"] for x in ua_arr]
        spider.slog.info(f"随机请求头中间件 RandomRequestUaMiddleware 已开启，生效脚本为: {spider.name}")

    def process_request(self, request, spider):
        # 根据权重来获取随机请求头 ua 信息
        request.headers.setdefault(b"User-Agent", self.get_random_ua_by_weight())



import hashlib
from pathlib import Path
from typing import TYPE_CHECKING, Any

import scrapy
from scrapy.http.request import NO_CALLBACK
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.python import to_bytes

from ayugespidertools.common.multiplexing import ReuseOperation
from ayugespidertools.common.utils import ToolsForAyu
from ayugespidertools.config import logger
from ayugespidertools.items import DataItem

__all__ = [
    "FilesDownloadPipeline",
    "files_download_by_scrapy",
]

if TYPE_CHECKING:
    from scrapy import Spider

    from ayugespidertools.common.typevars import DataItemModeStr


async def files_download_by_scrapy(
    spider: "Spider",
    file_path: str,
    file_url: str,
    item: Any,
    key: str,
    mode: "DataItemModeStr" = "namedtuple",
):
    request = scrapy.Request(file_url, callback=NO_CALLBACK)
    response = await maybe_deferred_to_future(spider.crawler.engine.download(request))
    if response.status != 200:
        return item

    headers_dict = ToolsForAyu.get_dict_form_scrapy_req_headers(
        scrapy_headers=response.headers
    )
    content_type = headers_dict.get("Content-Type")
    file_format = content_type.split("/")[-1].replace("jpeg", "jpg")
    file_guid = hashlib.sha1(to_bytes(file_url)).hexdigest()
    filename = f"{file_path}/{file_guid}.{file_format}"
    Path(filename).write_bytes(response.body)

    # Store file in item.
    if mode == "namedtuple":
        item[key] = DataItem(key_value=filename, notes=f"{key} 文件存储路径")
    else:
        item[key] = filename


class FilesDownloadPipeline:
    def __init__(self, file_path=None):
        self.file_path = file_path

    @classmethod
    def from_crawler(cls, crawler):
        _file_path = crawler.settings.get("FILES_STORE", None)
        assert _file_path is not None, "未配置 FILES_STORE 存储路径参数！"

        if not Path.exists(_file_path):
            logger.warning(f"FILES_STORE: {_file_path} 路径不存在，自动创建所需路径！")
            Path(_file_path).mkdir(parents=True)
        return cls(file_path=_file_path)

    async def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        judge_item = next(iter(item_dict.values()))
        file_url_keys = {
            key: value for key, value in item_dict.items() if key.endswith("_file_url")
        }
        if ReuseOperation.is_namedtuple_instance(judge_item):
            for key, value in file_url_keys.items():
                await files_download_by_scrapy(
                    spider, self.file_path, value.key_value, item, f"{key}_local"
                )
        else:
            for key, value in file_url_keys.items():
                await files_download_by_scrapy(
                    spider, self.file_path, value, item, f"{key}_local", "normal"
                )

        return item



# Define your TypeVar here
import threading
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Literal, NamedTuple, Optional, TypeVar, Union

from sqlalchemy import create_engine

if TYPE_CHECKING:
    import logging

    from loguru import Logger

    slogT = Union[Logger, logging.LoggerAdapter]

NoneType = type(None)
I_Str = TypeVar("I_Str", int, str)
B_Str = TypeVar("B_Str", bytes, str)
I_Str_N = TypeVar("I_Str_N", int, str, NoneType)
Str_Lstr = TypeVar("Str_Lstr", str, List[str])

AiohttpRequestMethodStr = Literal["GET", "POST"]
authMechanismStr = Literal[
    "SCRAM-SHA-1", "SCRAM-SHA-256", "MONGODB-CR", "MONGODB-X509", "PLAIN"
]
MysqlEngineStr = Literal["InnoDB", "MyISAM", "MEMORY", "NDB", "ARCHIVE"]
DataItemModeStr = Literal["normal", "namedtuple", "dict"]


class DatabaseSingletonMeta(type):
    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, engine_url, *args, **kwargs):
        if engine_url not in cls._instances:
            with cls._lock:
                if engine_url not in cls._instances:
                    instance = super().__call__(engine_url, *args, **kwargs)
                    cls._instances[engine_url] = instance
        return cls._instances[engine_url]


class DatabaseEngineClass(metaclass=DatabaseSingletonMeta):
    """database engine 单例模式：同一个 engine_url 只能存在一个实例"""

    def __init__(self, engine_url, *args, **kwargs):
        self.engine = create_engine(
            engine_url, pool_pre_ping=True, pool_recycle=3600 * 7, *args, **kwargs
        )


class MysqlConf(NamedTuple):
    host: str
    port: int
    user: str
    password: str
    database: str
    engine: MysqlEngineStr = "InnoDB"
    charset: str = "utf8mb4"
    collate: str = "utf8mb4_general_ci"
    odku_enable: bool = False


class MongoDBConf(NamedTuple):
    host: str = ""
    port: int = 27017
    user: str = ""
    password: str = ""
    database: Optional[str] = None
    authsource: Optional[str] = None
    authMechanism: authMechanismStr = "SCRAM-SHA-1"
    uri: Optional[str] = None


class PostgreSQLConf(NamedTuple):
    host: str
    port: int
    user: str
    password: str
    database: Optional[str] = None
    charset: str = "UTF8"


class ESConf(NamedTuple):
    hosts: str
    index_class: dict
    user: Optional[str] = None
    password: Optional[str] = None
    init: bool = False
    verify_certs: bool = False
    ca_certs: str = None
    client_cert: str = None
    client_key: str = None
    ssl_assert_fingerprint: str = None


class OracleConf(NamedTuple):
    host: str
    port: int
    user: str
    password: str
    service_name: Optional[str] = None
    encoding: str = "utf8"
    thick_lib_dir: Union[bool, str] = False


class AiohttpConf(NamedTuple):
    sleep: int
    proxy: str
    proxy_auth: str
    proxy_headers: dict
    retry_times: int
    limit: int
    ssl: bool
    verify_ssl: bool
    limit_per_host: int
    timeout: int
    allow_redirects: bool


class AlterItemTable(NamedTuple):
    """用于描述 AlterItem 中的 table 字段

    Attributes:
        name: table name
        notes: table name 的注释
    """

    name: str
    notes: str = ""


class AlterItem(NamedTuple):
    new_item: dict
    notes_dic: dict
    table: AlterItemTable


@dataclass
class AiohttpRequestArgs:
    url: Optional[str] = field(default=None)
    headers: Union[dict, None] = field(default=None)
    cookies: Union[dict, None] = field(default=None)
    method: AiohttpRequestMethodStr = field(default="GET")
    timeout: Union[int, None] = field(default=None)
    data: Union[dict, str, None] = field(default=None)
    proxy: Union[str, None] = field(default=None)
    proxy_auth: Union[str, None] = field(default=None)
    proxy_headers: Union[dict, None] = field(default=None)


class MQConf(NamedTuple):
    host: str
    port: int
    username: str
    password: str
    virtualhost: str = "/"
    heartbeat: int = 0
    socket_timeout: int = 1
    queue: Optional[str] = None
    durable: bool = True
    exclusive: bool = False
    auto_delete: bool = False
    exchange: Optional[str] = None
    routing_key: Optional[str] = None
    content_type: str = "text/plain"
    delivery_mode: int = 1
    mandatory: bool = True


class DynamicProxyConf(NamedTuple):
    proxy: str
    username: str
    password: str


class ExclusiveProxyConf(NamedTuple):
    proxy: str
    username: str
    password: str
    index: int


class KafkaConf(NamedTuple):
    bootstrap_servers: list
    topic: str
    key: str


class OssConf(NamedTuple):
    access_key: str
    access_secret: str
    endpoint: str
    bucket: str
    doc: Optional[str] = None
    upload_fields_suffix: Optional[str] = "_file_url"
    oss_fields_prefix: Optional[str] = "_"


class FieldAlreadyExistsError(Exception):
    def __init__(self, field_name: str):
        self.field_name = field_name
        self.message = f"字段 {field_name} 已存在！"
        super().__init__(self.message)


class EmptyKeyError(Exception):
    def __init__(self):
        self.message = "字段名不能为空！"
        super().__init__(self.message)



from ayugespidertools.scraper.middlewares.proxy.dynamic import (
    AbuDynamicProxyDownloaderMiddleware,
    DynamicProxyDownloaderMiddleware,
)
from ayugespidertools.scraper.middlewares.proxy.exclusive import (
    ExclusiveProxyDownloaderMiddleware,
)

__all__ = [
    "DynamicProxyDownloaderMiddleware",
    "ExclusiveProxyDownloaderMiddleware",
    "AbuDynamicProxyDownloaderMiddleware",
]



from typing import TYPE_CHECKING, Type, Union

from ayugespidertools.common.multiplexing import ReuseOperation

try:
    from elasticsearch_dsl import Document, connections
except ImportError:
    # pip install elasticsearch-dsl
    pass

__all__ = ["AyuESPipeline", "dynamic_es_document"]

if TYPE_CHECKING:
    from ayugespidertools.common.typevars import ESConf

    DocumentType = Union[Type[Document], type]


def dynamic_es_document(class_name, fields, index_settings=None):
    class_attrs = fields.copy()
    if index_settings:
        class_attrs["Index"] = type("Index", (), index_settings)

    return type(class_name, (Document,), class_attrs)


class AyuESPipeline:
    es_conf: "ESConf"
    es_type: "DocumentType"

    def open_spider(self, spider):
        assert hasattr(spider, "es_conf"), "未配置 elasticsearch 连接信息！"
        self.es_conf = spider.es_conf
        _hosts_lst = self.es_conf.hosts.split(",")
        if any([self.es_conf.user is not None, self.es_conf.password is not None]):
            http_auth = (self.es_conf.user, self.es_conf.password)
        else:
            http_auth = None
        connections.create_connection(
            hosts=_hosts_lst,
            http_auth=http_auth,
            verify_certs=self.es_conf.verify_certs,
            ca_certs=self.es_conf.ca_certs,
            client_cert=self.es_conf.client_cert,
            client_key=self.es_conf.client_key,
            ssl_assert_fingerprint=self.es_conf.ssl_assert_fingerprint,
        )

    def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        alert_item = ReuseOperation.reshape_item(item_dict)
        if not (new_item := alert_item.new_item):
            return

        if not hasattr(self, "es_type"):
            fields_define = {k: v.notes for k, v in item_dict.items()}
            index_define = self.es_conf.index_class
            index_define["name"] = alert_item.table.name
            self.es_type = dynamic_es_document("ESType", fields_define, index_define)
            if self.es_conf.init:
                self.es_type.init()
        es_item = self.es_type(**new_item)
        es_item.save()
        return item



import itertools
import math
import random
from typing import TYPE_CHECKING, Any, Optional, Tuple, Union

try:
    import cv2
    import numpy as np
except ImportError:
    # pip install ayugespidertools[all]
    pass

__all__ = ["CvnpilKit", "BezierTrajectory"]

if TYPE_CHECKING:
    from cv2.typing import MatLike


class CvnpilKit:
    @staticmethod
    def get_array_dimension(array: Union[frozenset, list, set, tuple]) -> int:
        """获取 array 的维度

        Args:
            array: 数组

        Returns:
            1). 层级数
        """
        # 其实直接返回 len(array) 即可
        return len(np.array(array).shape)

    @staticmethod
    def read_image_data(
        img: Union[bytes, str],
        flags: int = cv2.IMREAD_COLOR,
    ) -> np.ndarray:
        """
        用 opencv 读取图片数据
        Args:
            img: 图片参数，类型需要是路径 str 或 bytes 数据
            flags: 应用颜色通道类型

        Returns:
            img_cv: opencv 读取背景图片的数据
        """
        assert type(img) in [str, bytes], "img 参数需要是路径或 bytes 数据"
        if isinstance(img, bytes):
            img_buf = np.frombuffer(img, np.uint8)
            img_cv = cv2.imdecode(img_buf, cv2.IMREAD_ANYCOLOR)
        else:
            # 读取图片，读进来直接是 BGR 格式数据格式在 0~255
            img_cv = cv2.imread(img, flags)
        return img_cv

    @classmethod
    def clear_white(cls, img):
        """清除图片的空白区域，这里主要清除滑块的空白

        Args:
            img: 待处理的图片

        Returns:
            1). 清除图片空白区域的图片
        """
        rows, cols, channel = img.shape
        min_x = 255
        min_y = 255
        max_x = 0
        max_y = 0
        for x, y in itertools.product(range(1, rows), range(1, cols)):
            t = set(img[x, y])
            if len(t) >= 2:
                if x <= min_x:
                    min_x = x
                elif x >= max_x:
                    max_x = x

                if y <= min_y:
                    min_y = y
                elif y >= max_y:
                    max_y = y
        return img[min_x:max_x, min_y:max_y]

    @classmethod
    def image_edge_detection(cls, img):
        """图像边缘检测处理，识别图片边缘

        Args:
            img: 需要处理的图片，用于边缘检测使用

        Returns:
            1). 处理后的图片
        """
        return cv2.Canny(img, 100, 200)

    @classmethod
    def _template_match(
        cls,
        bg: "MatLike",
        slider: "MatLike",
        out: Optional[str] = None,
    ) -> int:
        """模板匹配找出滑块缺口的距离

        Args:
            bg: 背景图片
            slider: 缺口(滑块)图片
            out: 展示图片的存储全路径，会将绘制的图片保存至此

        Returns:
            tl[0]: 滑块缺口的距离
        """
        res = cv2.matchTemplate(bg, slider, cv2.TM_CCOEFF_NORMED)
        # 寻找最优匹配(一维数组当作向量,用 Mat 定义) 中最小值和最大值的位置
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
        # 左上角点的坐标
        tl = max_loc

        # 是否要输出绘制图像
        if out:
            # 绘制方框
            th, tw = slider.shape[:2]
            # 右下角点的坐标
            br = (tl[0] + tw, tl[1] + th)
            cv2.rectangle(bg, tl, br, (0, 0, 255), 2)
            cv2.imwrite(out, bg)
        # 返回缺口的X坐标
        return tl[0]

    @classmethod
    def discern_gap(
        cls,
        bg: Union[str, bytes],
        slider: Union[str, bytes],
        out: Optional[str] = None,
    ):
        """识别滑块缺口方法

        Args:
            bg: 带缺口的背景图，可以是全路径图片，也可以是图片的 bytes 数据
            slider: 滑块图，可以是全路径图片，也可以是图片的 bytes 数据
            out: 绘制图展示的存储地址，参数格式为图片的全路径

        Returns:
            1): 滑块缺口横坐标
        """
        slider_cv = CvnpilKit.read_image_data(slider)
        bg_cv = CvnpilKit.read_image_data(bg, cv2.IMREAD_GRAYSCALE)
        slider_clr = cls.clear_white(slider_cv)
        slider_clr = cv2.cvtColor(slider_clr, cv2.COLOR_RGB2GRAY)
        slider_img = cls.image_edge_detection(slider_clr)
        bg_img = cls.image_edge_detection(bg_cv)
        slider_img = cv2.cvtColor(slider_img, cv2.COLOR_GRAY2RGB)
        bg_img = cv2.cvtColor(bg_img, cv2.COLOR_GRAY2RGB)
        # 输出横坐标, 即滑块在图片上的位置
        return cls._template_match(bg_img, slider_img, out)

    @classmethod
    def identify_gap(
        cls,
        bg: Union[bytes, str],
        slider: Union[bytes, str],
        out: Optional[str] = None,
    ) -> int:
        """通过背景图片和缺口图片识别出滑块距离

        Args:
            bg: 背景图片，可以是图片的全路径，也可以是图片的 bytes 内容
            slider: 缺口（滑块）图片，可以是图片的全路径，也可以是图片的 bytes 内容
            out: 输出图片路径，示例：doc/test.jpg；此参数如果为空，则不输出标记后的图片

        Returns:
            tl[0]: 滑块缺口距离
        """
        # 读取图片数据
        bg_cv = cls.read_image_data(bg)
        slider_cv = cls.read_image_data(slider, cv2.IMREAD_GRAYSCALE)
        # 识别图片边缘
        bg_edge = cv2.Canny(bg_cv, 100, 200)
        slider_edge = cv2.Canny(slider_cv, 100, 200)
        # 转换图片格式
        bg_img = cv2.cvtColor(bg_edge, cv2.COLOR_GRAY2RGB)
        slider_img = cv2.cvtColor(slider_edge, cv2.COLOR_GRAY2RGB)
        return cls._template_match(bg_img, slider_img, out)

    @staticmethod
    def match_gap(bg: Union[str, bytes], slider: Union[str, bytes]):
        """滑块坐标定位方法

        Args:
            bg: 1) 带缺口的背景图，全路径; 2) 或者是背景图的 bytes 数据
            slider: 1） 滑块图，全路径; 2）或者是滑块图的 bytes 数据

        Returns:
            loc[1][0]: 滑块位置坐标
        """
        bg_cv = CvnpilKit.read_image_data(bg, cv2.IMREAD_GRAYSCALE)
        slider_cv = CvnpilKit.read_image_data(slider, cv2.IMREAD_GRAYSCALE)

        run = 1
        # 这里的返回值根据不同版本的 open-cv 其返回结果会有不同的个数
        # w, h, z = slider_cv.shape[::-1]
        # 在背景图里面查找滑块图的位置
        res = cv2.matchTemplate(bg_cv, slider_cv, cv2.TM_CCOEFF_NORMED)
        # 使用二分法查找阈值的精确值
        lft: float = 0.0
        rgt: float = 1.0
        loc: Tuple[np.ndarray[Any, np.dtype[np.signedinteger[Any]]], ...] = ()
        while run < 20:
            run += 1
            threshold = (rgt + lft) / 2
            if threshold < 0:
                # 逻辑走到这里，则说明出错了，并未找出目标位置
                return None

            # 匹配程度大于百分之 threshold 的坐标 x, y
            loc = np.where(res >= threshold)
            if len(loc[1]) > 1:
                rgt += (rgt - lft) / 2
            elif len(loc[1]) == 1:
                # 找到目标区域起点x坐标为：loc[1][0]
                break
            elif len(loc[1]) < 1:
                rgt -= (rgt - lft) / 2
        # 返回 x 坐标
        return loc[1][0]

    @staticmethod
    def get_normal_track(space):
        """通用的根据滑块距离获取轨迹数组方法

        Args:
            space: 滑块缺口距离

        Returns:
            tracks_list: 生成的轨迹数组
        """
        x = [0, 0]
        y = [0, 0, 0]
        z = [0]
        # x
        count = np.linspace(-math.pi / 2, math.pi / 2, random.randrange(20, 30))
        func = list(map(math.sin, count))
        nx = [i + 1 for i in func]
        add = random.randrange(10, 15)
        sadd = space + add
        x.extend(list(map(lambda x: x * (sadd / 2), nx)))
        # x.extend(np.linspace(sadd, space, 4 if add > 12 else 3))
        x.extend(np.linspace(sadd, space, 3 if add > 12 else 2))
        x = [math.floor(i) for i in x]
        # y
        for i in range(len(x) - 2):
            if y[-1] < 30:
                y.append(y[-1] + random.choice([0, 0, 1, 1, 2, 2, 1, 2, 0, 0, 3, 3]))
            else:
                y.append(
                    y[-1] + random.choice([0, 0, -1, -1, -2, -2, -1, -2, 0, 0, -3, -3])
                )
        # z
        for i in range(len(x) - 1):
            # z.append((z[-1] // 100 * 100) + 100 + random.choice([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2]))
            z.append(
                (z[-1] // 100 * 100)
                + 100
                + random.choice([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3])
            )

        tracks_list = list(map(list, zip(x, y, z)))
        tracks_list = [x for x in tracks_list if x[0] > 0]
        return tracks_list


class BezierTrajectory:
    """贝塞尔曲线轨迹生成器

    Examples:
        >>> bt = BezierTrajectory()
        >>> gen_data = bt.gen_track(start=[50, 268], end=[367, 485], num=45, order=4, mode=2)
        >>> track = gen_data["trackArray"]
    """

    def _generate_control_points(self, track: list):
        """计算贝塞尔曲线的控制点"""
        track_len = len(track)

        def calculate_bezier_point(x):
            t = (x - track[0][0]) / (track[-1][0] - track[0][0])
            y = np.array([0, 0], dtype=np.float64)
            for s in range(len(track)):
                y += track[s] * (
                    (
                        math.factorial(track_len - 1)
                        / (math.factorial(s) * math.factorial(track_len - 1 - s))
                    )
                    * math.pow(t, s)
                    * math.pow((1 - t), track_len - 1 - s)
                )
            return y[1]

        return calculate_bezier_point

    def _type(self, mode, x, length):
        numbers = []
        pin = (x[1] - x[0]) / length
        if mode == 0:
            numbers.extend(i * pin for i in range(length))
            if pin >= 0:
                numbers = numbers[::-1]
        elif mode == 1:
            for i in range(length):
                numbers.append(1 * ((i * pin) ** 2))
            numbers = numbers[::-1]
        elif mode == 2:
            for i in range(length):
                numbers.append(1 * ((i * pin - x[1]) ** 2))

        elif mode == 3:
            track = [
                np.array([0, 0]),
                np.array([(x[1] - x[0]) * 0.8, (x[1] - x[0]) * 0.6]),
                np.array([x[1] - x[0], 0]),
            ]
            fun = self._generate_control_points(track)
            numbers = [0]
            numbers.extend(fun(i * pin) + numbers[-1] for i in range(1, length))
            if pin >= 0:
                numbers = numbers[::-1]
        numbers = np.abs(np.array(numbers) - max(numbers))
        normal_numbers = (
            (numbers - numbers[numbers.argmin()])
            / (numbers[numbers.argmax()] - numbers[numbers.argmin()])
        ) * (x[1] - x[0]) + x[0]
        normal_numbers[0] = x[0]
        normal_numbers[-1] = x[1]
        return normal_numbers

    def simulation(self, start, end, order=1, deviation=0, bias=0.5):
        """模拟贝塞尔曲线的绘制过程

        Args:
            start: 开始点的坐标
            end: 结束点的坐标
            order: 几阶贝塞尔曲线，越大越复杂
            deviation: 轨迹上下波动的范围
            bias: 波动范围的分布位置

        Returns:
            1). 返回一个字典 equation 对应该曲线的方程，P 对应贝塞尔曲线的影响点
        """
        start = np.array(start)
        end = np.array(end)
        shake_num = []
        if order != 1:
            e = (1 - bias) / (order - 1)
            shake_num = [[bias + e * i, bias + e * (i + 1)] for i in range(order - 1)]

        track_lst = [start]

        t = random.choice([-1, 1])
        w = 0
        for i in shake_num:
            px1 = start[0] + (end[0] - start[0]) * (
                random.random() * (i[1] - i[0]) + (i[0])
            )
            p = np.array(
                [px1, self._generate_control_points([start, end])(px1) + t * deviation]
            )
            track_lst.append(p)
            w += 1
            if w >= 2:
                w = 0
                t = -1 * t

        track_lst.append(end)
        return {
            "equation": self._generate_control_points(track_lst),
            "P": np.array(track_lst),
        }

    def gen_track(
        self,
        start: Union[np.ndarray, list],
        end: Union[np.ndarray, list],
        num: int,
        order: int = 1,
        deviation: int = 0,
        bias=0.5,
        mode=0,
        shake_num=0,
        yhh=10,
    ) -> dict:
        """生成轨迹数组

        Args:
            start: 开始点的坐标
            end: 结束点的坐标
            num: 返回的数组的轨迹点的数量
            order: 几阶贝塞尔曲线，越大越复杂
            deviation: 轨迹上下波动的范围
            bias: 波动范围的分布位置
            mode: 0 表示均速滑动，1 表示先慢后快，2 表示先快后慢，3 表示先慢中间快后慢
            shake_num: 在终点来回摆动的次数
            yhh: 在终点来回摆动的范围

        Returns:
            1). 返回一个字典 trackArray 对应轨迹数组，P 对应贝塞尔曲线的影响点
        """
        s: list = []
        fun = self.simulation(start, end, order, deviation, bias)
        w = fun["P"]
        fun = fun["equation"]
        if shake_num != 0:
            track_number = round(num * 0.2 / (shake_num + 1))
            num -= num * (shake_num + 1)

            x_track_array = self._type(mode, [start[0], end[0]], num)
            s.extend([i, fun(i)] for i in x_track_array)
            dq = yhh / shake_num
            kg = 0
            ends = np.copy(end)
            for i in range(shake_num):
                if kg == 0:
                    d = np.array(
                        [
                            end[0] + (yhh - dq * i),
                            ((end[1] - start[1]) / (end[0] - start[0]))
                            * (end[0] + (yhh - dq * i))
                            + (
                                end[1]
                                - ((end[1] - start[1]) / (end[0] - start[0])) * end[0]
                            ),
                        ]
                    )
                    kg = 1
                else:
                    d = np.array(
                        [
                            end[0] - (yhh - dq * i),
                            ((end[1] - start[1]) / (end[0] - start[0]))
                            * (end[0] - (yhh - dq * i))
                            + (
                                end[1]
                                - ((end[1] - start[1]) / (end[0] - start[0])) * end[0]
                            ),
                        ]
                    )
                    kg = 0
                y = self.gen_track(
                    ends,
                    d,
                    track_number,
                    order=2,
                    deviation=0,
                    bias=0.5,
                    mode=0,
                    shake_num=0,
                    yhh=10,
                )
                s += list(y["trackArray"])
                ends = d
            y = self.gen_track(
                ends,
                end,
                track_number,
                order=2,
                deviation=0,
                bias=0.5,
                mode=0,
                shake_num=0,
                yhh=10,
            )
            s += list(y["trackArray"])

        else:
            x_track_array = self._type(mode, [start[0], end[0]], num)
            s.extend([i, fun(i)] for i in x_track_array)
        return {"trackArray": np.array(s), "P": w}



import asyncio
from typing import TYPE_CHECKING

import aiomysql
from scrapy.utils.defer import deferred_from_coro

from ayugespidertools.common.expend import MysqlPipeEnhanceMixin
from ayugespidertools.common.multiplexing import ReuseOperation

__all__ = [
    "AyuAsyncMysqlPipeline",
]

if TYPE_CHECKING:
    from ayugespidertools.common.typevars import MysqlConf, slogT


class AyuAsyncMysqlPipeline(MysqlPipeEnhanceMixin):
    mysql_conf: "MysqlConf"
    slog: "slogT"
    pool: aiomysql.Pool
    running_tasks: set

    def open_spider(self, spider):
        assert hasattr(spider, "mysql_conf"), "未配置 Mysql 连接信息！"
        self.running_tasks = set()
        self.slog = spider.slog
        self.mysql_conf = spider.mysql_conf
        return deferred_from_coro(self._open_spider(spider))

    async def _open_spider(self, spider):
        self.pool = await aiomysql.create_pool(
            host=self.mysql_conf.host,
            port=self.mysql_conf.port,
            user=self.mysql_conf.user,
            password=self.mysql_conf.password,
            db=self.mysql_conf.database,
            charset=self.mysql_conf.charset,
            cursorclass=aiomysql.DictCursor,
            autocommit=True,
        )

    async def insert_item(self, item_dict):
        async with self.pool.acquire() as conn:
            async with conn.cursor() as cursor:
                alter_item = ReuseOperation.reshape_item(item_dict)
                new_item = alter_item.new_item
                sql, args = self._get_sql_by_item(
                    table=alter_item.table.name,
                    item=new_item,
                    odku_enable=self.mysql_conf.odku_enable,
                )
                await cursor.execute(sql, args)

    async def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        task = asyncio.create_task(self.insert_item(item_dict))
        self.running_tasks.add(task)
        await task
        task.add_done_callback(lambda t: self.running_tasks.discard(t))
        return item

    async def _close_spider(self):
        await self.pool.wait_closed()

    def close_spider(self, spider):
        self.pool.close()
        return deferred_from_coro(self._close_spider())



import hashlib
from typing import TYPE_CHECKING, Any

import scrapy
from scrapy.http.request import NO_CALLBACK
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.python import to_bytes

from ayugespidertools.common.multiplexing import ReuseOperation
from ayugespidertools.common.utils import ToolsForAyu
from ayugespidertools.extras.oss import AliOssBase
from ayugespidertools.items import DataItem

__all__ = ["AyuAsyncOssPipeline"]

if TYPE_CHECKING:
    from scrapy import Spider

    from ayugespidertools.common.typevars import OssConf


async def files_download_by_scrapy(
    spider: "Spider",
    file_url: str,
    item: Any,
):
    request = scrapy.Request(file_url, callback=NO_CALLBACK)
    response = await maybe_deferred_to_future(spider.crawler.engine.download(request))
    if response.status != 200:
        return item

    headers_dict = ToolsForAyu.get_dict_form_scrapy_req_headers(
        scrapy_headers=response.headers
    )
    content_type = headers_dict.get("Content-Type")
    file_format = content_type.split("/")[-1].replace("jpeg", "jpg")
    file_guid = hashlib.sha1(to_bytes(file_url)).hexdigest()
    filename = f"{file_guid}.{file_format}"
    return response, filename


class AyuAsyncOssPipeline:
    oss_bucket: AliOssBase
    oss_conf: "OssConf"

    def open_spider(self, spider):
        assert hasattr(spider, "oss_conf"), "未配置 oss 参数！"
        self.oss_conf = spider.oss_conf
        oss_conf_dict = self.oss_conf._asdict()
        self.oss_bucket = AliOssBase(**oss_conf_dict)

    async def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        judge_item = next(iter(item_dict.values()))
        file_url_keys = {
            key: value
            for key, value in item_dict.items()
            if key.endswith(self.oss_conf.upload_fields_suffix)
        }

        if ReuseOperation.is_namedtuple_instance(judge_item):
            for key, value in file_url_keys.items():
                file_url = value.key_value
                r, filename = await files_download_by_scrapy(spider, file_url, item)

                self.oss_bucket.put_oss(put_bytes=r.body, file=filename)
                item[f"{self.oss_conf.oss_fields_prefix}{key}"] = DataItem(
                    key_value=filename, notes=f"{key} 对应的 oss 存储字段"
                )

        else:
            for key, value in file_url_keys.items():
                r, filename = await files_download_by_scrapy(spider, value, item)

                self.oss_bucket.put_oss(put_bytes=r.body, file=filename)
                item[f"{self.oss_conf.oss_fields_prefix}{key}"] = filename

        return item



from typing import TYPE_CHECKING, Optional, Tuple

from gridfs import GridFS
from pymongo import MongoClient

from ayugespidertools.common.typevars import DatabaseSingletonMeta

__all__ = [
    "MongoDbBase",
    "MongoDBEngineClass",
]

if TYPE_CHECKING:
    import pymongo
    from pymongo import database

    from ayugespidertools.common.typevars import authMechanismStr


class MongoDBEngineClass(metaclass=DatabaseSingletonMeta):
    """mongodb 连接单例模式：同一个 engine_url 只能存在一个 conn 实例"""

    def __init__(self, engine_url, *args, **kwargs):
        self.engine = MongoDbBase.connects(uri=engine_url)


class MongoDbBase:
    """mongodb 数据库的相关操作"""

    @staticmethod
    def connects(
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: str = "localhost",
        port: int = 27017,
        authsource: str = "admin",
        authMechanism: "authMechanismStr" = "SCRAM-SHA-1",
        database: Optional[str] = None,
        uri: Optional[str] = None,
    ) -> Tuple["pymongo.MongoClient", "database.Database"]:
        """初始化 mongo 连接句柄
        可传入 user, password, host 等参数的形式，也可只传入 uri 的方式

        Args:
            user: 用户名
            password: 用户对应的密码
            host: mongoDB 链接需要的 host
            port: mongoDB 链接需要的端口
            authsource: mongoDB 身份验证需要的数据库名称
            authMechanism: mongoDB 身份验证机制
            database: mongoDB 链接需要的数据库
            uri: mongoDB uri，需要包含 database 参数， demo: 'mongodb://host/my_database'
        """
        if uri is not None:
            conn = MongoClient(uri)
            db = conn.get_database()

        else:
            conn = MongoClient(
                host=host,
                port=port,
                username=user,
                password=password,
                authSource=authsource,
                authMechanism=authMechanism,
            )
            db = conn[database]
        return conn, db

    @classmethod
    def getFileMd5(cls, db, _id, collection):
        gridfs_col = GridFS(db, collection)
        gf = gridfs_col.get(_id)
        md5 = gf.md5
        _id = gf._id
        return {"_id": _id, "md5": md5}

    @classmethod
    def upload(cls, db, file_name, _id, content_type, collection, file_data):
        """上传文件

        Args:
            db: 目标库对应连接
            file_name: 上传至 mongoDB 的 GridFS 存储桶里的文件名
            _id: 唯一 id，雪花 id，用来标识此上传任务和图片的唯一
            content_type: 上传文件的类型，示例：image/jpeg
            collection: 存储至 GridFS 存储桶名称
            file_data: 文件的 bytes 内容

        Returns:
            gridfs_id: 上传至 GridFS 上的文件 ID 标识
            image_id: 上传至 GridFS 上的文件 MD5 标识
        """
        metadata = {
            "_contentType": content_type,
            "isThumb": "true",
            "targetId": _id,
            "_class": "com.ccr.dc.admin.mongo.MongoFsMetaData",
        }

        gridfs_col = GridFS(db, collection)
        # 当存储桶中不存在此文件，才需要上传
        if not gridfs_col.exists(filename=file_name):
            gridfs_id = gridfs_col.put(
                data=file_data,
                content_type=content_type,
                filename=file_name,
                metadata=metadata,
            )
            md5 = cls.getFileMd5(db, gridfs_id, "fs")["md5"]
            image_id = f"/file/find/{str(gridfs_id)}/{md5}"
            return gridfs_id, image_id

        # 否则，只需返回文件的 id 等标识即可
        res = gridfs_col.find_one({"filename": file_name})
        return res._id, f"/file/find/{str(res._id)}/{res.md5}"



import asyncio
import urllib.parse
from typing import TYPE_CHECKING

import motor.motor_asyncio

from ayugespidertools.common.mongodbpipe import AsyncioAsynchronous
from ayugespidertools.common.multiplexing import ReuseOperation

__all__ = ["AyuAsyncMongoPipeline"]

if TYPE_CHECKING:
    from motor.core import AgnosticClient, AgnosticDatabase


class AyuAsyncMongoPipeline:
    client: "AgnosticClient"
    db: "AgnosticDatabase"

    def open_spider(self, spider):
        assert hasattr(spider, "mongodb_conf"), "未配置 MongoDB 连接信息！"
        if spider.mongodb_conf.uri is not None:
            _mongo_uri = spider.mongodb_conf.uri
        else:
            _encoded_pwd = urllib.parse.quote_plus(spider.mongodb_conf.password)
            _mongo_uri = (
                f"mongodb://{spider.mongodb_conf.user}:{_encoded_pwd}"
                f"@{spider.mongodb_conf.host}:{spider.mongodb_conf.port}/"
                f"{spider.mongodb_conf.database}?authSource={spider.mongodb_conf.authsource}"
                f"&authMechanism={spider.mongodb_conf.authMechanism}",
            )

        self.client = motor.motor_asyncio.AsyncIOMotorClient(_mongo_uri)
        self.db = self.client.get_database()

    async def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        await asyncio.shield(
            AsyncioAsynchronous().process_item_template(
                item_dict=item_dict,
                db=self.db,
            )
        )
        return item

    def close_spider(self, spider):
        self.client.close()



import json
import random
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Literal, Optional, Union
from urllib.parse import urlparse

import requests

from ayugespidertools.common.encryption import EncryptOperation
from ayugespidertools.common.multiplexing import ReuseOperation
from ayugespidertools.common.params import Param
from ayugespidertools.config import logger
from ayugespidertools.extras.ext import AppConfManageMixin
from ayugespidertools.formatdata import DataHandle

__all__ = ["ToolsForAyu"]

if TYPE_CHECKING:
    from scrapy.http import Response
    from scrapy.http.headers import Headers

    from ayugespidertools.common.typevars import Str_Lstr

RemoteFormatStr = Literal["json", "hcl", "yaml", "xml"]
RemoteTypeStr = Literal["consul", "nacos"]
RemoteConfNameStr = Literal[
    "mongodb",
    "mongodb:uri",
    "mysql",
    "postgresql",
    "elasticsearch",
    "oracle",
    "oss:ali",
    "rabbitmq",
    "kafka",
    "dynamicproxy",
    "exclusiveproxy",
]


class ToolsForAyu(AppConfManageMixin):
    """这里用于存放框架所依赖的方法"""

    @classmethod
    @lru_cache(maxsize=16)
    def get_remote_kvs(
        cls,
        url: str,
        remote_type: RemoteTypeStr = "consul",
        token: Optional[str] = None,
    ) -> str:
        """获取远程配置中的 key_values 信息

        Args:
            url: 获取 kvs 所需的 url
            remote_type: 配置类型
            token: consul token or nacos token; nacos 的 token 值直接在 url 中构造即可

        Returns:
            1). 远程配置中 key_values 的详细信息
        """
        headers = {"X-Consul-Token": token} if remote_type == "consul" else None
        try:
            r = requests.get(
                url,
                headers=headers,
                verify=False,
                timeout=(
                    Param.requests_req_timeout,
                    Param.requests_res_timeout,
                ),
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
        ) as e:
            raise ValueError(f"请求远程配置 {remote_type} api 超时!") from e

        url_params = urlparse(url).query
        if remote_type == "consul":
            if "raw" in url_params:
                return r.text
            return EncryptOperation.base64_decode(decode_data=r.json()[0]["Value"])
        return r.text

    @classmethod
    def fetch_remote_conf(
        cls,
        conf_name: RemoteConfNameStr,
        url: str,
        format: RemoteFormatStr = "json",
        remote_type: RemoteTypeStr = "consul",
        token: Optional[str] = None,
    ) -> dict:
        """获取远程中的项目配置信息

        Args:
            conf_name: 需要获取的配置
            token: consul token or nacos token
            format: 远程配置中的格式，默认为 json 格式
            remote_type: 配置类型
            url: 获取远程配置所需的 url

        Returns:
            1). 远程应用配置中心中的配置信息（key 值为小写）
        """
        conf_value = cls.get_remote_kvs(url, remote_type, token)
        if format == "json":
            conf_data = json.loads(conf_value)
        elif format == "xml":
            conf_data = cls.xml_parser(conf_value)
        elif format == "yaml":
            conf_data = cls.yaml_parser(conf_value)
        elif format == "hcl":
            conf_data = cls.hcl_parser(conf_value)
        else:
            raise ValueError(f"{conf_name} 暂不支持该格式的配置")

        _conf = conf_data.get(conf_name, {})
        if not _conf:
            logger.info(f"远程配置 {remote_type} 中未设置 {conf_name}")
        return _conf

    @classmethod
    @DataHandle.simple_deal_for_extract
    def extract_with_css(
        cls,
        response: "Response",
        query: str,
        get_all: bool = False,
        return_selector: bool = False,
    ):
        """使用 scrapy 的 css 提取信息

        Args:
            response: scrapy response 或者是 selector 对象
            query: css 提取的规则
            get_all: 提取模式，默认：False, 提取符合中的一个
            return_selector: 是否返回选择器对象

        Returns:
            1). 提取的内容
        """
        if return_selector:
            return response.css(query)
        if get_all:
            return response.css(query).getall()
        else:
            return response.css(query).get(default="").strip()

    @classmethod
    @DataHandle.simple_deal_for_extract
    def extract_with_xpath(
        cls,
        response: "Response",
        query: str,
        get_all: bool = False,
        return_selector: bool = False,
    ):
        """使用 scrapy 的 xpath 提取信息

        Args:
            response: scrapy response 或者是 selector 对象
            query: css 提取的规则
            get_all: 提取模式，默认：False, 提取符合中的一个
            return_selector: 是否返回选择器对象

        Returns:
            1). 提取的内容
        """
        if return_selector:
            return response.xpath(query)
        if get_all:
            return response.xpath(query).getall()
        else:
            return response.xpath(query).get(default="").strip()

    @classmethod
    @DataHandle.simple_deal_for_extract
    def extract_with_json(cls, json_data: dict, query: Union[str, List[str]]):
        """scrapy 中提取 json 数据遇到的情况

        Args:
            json_data: scrapy response 响应内容中的 json 格式数据
            query: json 提取的规则

        Returns:
            1). 提取的内容
        """
        # 如果输入的提取规则参数的格式为字符；或者参数格式是个列表，但是只含有一个元素的情况时
        if any(
            [isinstance(query, str), all([isinstance(query, list), len(query) == 1])]
        ):
            if isinstance(query, str):
                return json_data.get(query, "")
            return json_data.get(query[0], "")

        # 循环取值时的处理
        for curr_q in query:
            # 这里循环时不对 json_data 的类型做判断，如果此时 json_data 类型无 get 方法，不处理
            json_data = json_data.get(curr_q, "")
            if not json_data:
                return json_data
        return json_data

    @classmethod
    def extract_with_json_rules(cls, json_data: dict, query_rules: List["Str_Lstr"]):
        """当提取 json 某个数据时，可以在某些字段中取值，只要返回其中任意一个含有数据的值即可

        Args:
            json_data: scrapy response 响应内容中的 json 格式数据
            query_rules: json 提取的规则列表

        Returns:
            1). 提取的内容
        """
        # 先判断层级，最多为 2 层
        depth_num = ReuseOperation.get_array_depth(query_rules)
        assert depth_num <= 2, "query_rules 参数错误，请输入深度最多为 2 的参数！"

        for query in query_rules:
            if extract_res := cls.extract_with_json(json_data=json_data, query=query):
                return extract_res
        return ""

    @staticmethod
    def first_not_none(data_lst: List[Any]) -> Any:
        """获取列表中第一个不为 None 的值

        Args:
            data_lst: 数据列表

        Returns:
            1). 第一个不为 None 的值
        """
        _res = [x for x in data_lst if x is not None]
        return _res[0] if _res else None

    @staticmethod
    def get_dict_form_scrapy_req_headers(scrapy_headers: "Headers") -> dict:
        """根据 scrapy request 中的 headers 信息转化为正常的 dict 格式

        Args:
            scrapy_headers: scrapy 的 request headers 内容

        Returns:
            1). 转化 dict 后的 headers 内容
        """
        return {
            str(b_key, encoding="utf-8"): str(b_value_list[0], encoding="utf-8")
            for b_key, b_value_list in dict(scrapy_headers).items()
        }

    @classmethod
    def get_data_urls_by_img(cls, mediatype: str, data: Union[bytes, str]) -> str:
        """根据本地、远程或 bytes 内容的图片生成 Data URLs 格式的数据
        Data URLs 格式示例:
            data:image/png;base64,iVB...
            data:text/html,%3Ch1%3EHello%2C%20World%21%3C%2Fh1%3E
        关于 Data URLs 更多的描述，其参考文档: https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/Data_URLs

        Args:
            mediatype: MIME 类型字符串，例如 'image/jpeg' JPEG 图像文件。
                如果省略，则默认为 text/plain;charset=US-ASCII
            data: 用于获取其 base64 编码的二进制数据
                参数格式可以为全路径图片，或 bytes 内容

        Returns:
            1). Data URLs 格式数据
        """
        assert type(data) in [
            str,
            bytes,
        ], "图片转 Data URLs 的参数 data 需要是全路径 str 或 bytes 数据"

        if isinstance(data, str):
            data_bytes = Path(data).read_bytes()
            data_base64_encoded = EncryptOperation.base64_encode(encode_data=data_bytes)

        else:
            data_base64_encoded = EncryptOperation.base64_encode(encode_data=data)
        return f"data:image/{mediatype};base64,{data_base64_encoded}"

    @staticmethod
    def gen_selenium_track(distance):
        """最简陋的根据缺口距离获取轨迹的方法，以供 selenium 使用

        Args:
            distance: 滑块缺口的距离

        Returns:
            tracks_dict:
                forward_tracks: 往右划的轨迹数组
                back_tracks: 回退（往左划）的轨迹数组
        """
        distance += 20
        v = 0
        t = 0.2
        forward_tracks = []
        current = 0
        mid = distance * 3 / 5
        while current < distance:
            a = 2 if current < mid else -3
            s = v * t + 0.5 * a * (t**2)
            v = v + a * t
            current += s
            forward_tracks.append(round(s))

        back_tracks = [-3, -3, -2, -2, -2, -2, -2, -1, -1, -1]
        return {"forward_tracks": forward_tracks, "back_tracks": back_tracks}

    @staticmethod
    def gen_tracks(distance):
        """轨迹生成方法

        Args:
            distance: 滑块缺口的距离

        Returns:
            xyt: 轨迹数组
        """
        t_list = [random.randint(50, 160)]
        x_list = [random.randint(5, 11)]
        y_list = []
        # 生成x坐标轨迹, 生成t坐标轨迹
        for j in range(1, distance):
            x_list.append(x_list[j - 1] + random.randint(2, 4))
            if x_list[j] > distance:
                break

        diff = x_list[-1] - distance
        for j in range(diff):
            x_list.append(x_list[-1] + random.randint(-2, -1))
            if x_list[-1] <= distance:
                x_list[-1] = distance
                break

        length = len(x_list)
        # 生成y坐标轨迹
        for i in range(1, length + 1):
            if i < int(length * 0.4):
                y_list.append(0)
            elif i < int(length * 0.65):
                y_list.append(-1)
            elif i < int(length * 0.77):
                y_list.append(-2)
            elif i < int(length * 0.95):
                y_list.append(-3)
            else:
                y_list.append(-4)
            t_list.append(t_list[i - 1] + random.randint(20, 80))

        # 生成t的坐标
        xyt = list(zip(x_list, y_list, t_list))
        for j in range(length):
            xyt[j] = list(xyt[j])
        return xyt



import json
from typing import Optional

from kafka import KafkaProducer
from kafka.errors import KafkaError

from ayugespidertools.common.multiplexing import ReuseOperation
from ayugespidertools.config import logger

__all__ = ["AyuKafkaPipeline"]


class KafkaProducerClient:
    def __init__(self, bootstrap_servers: list) -> None:
        self.producer = KafkaProducer(
            bootstrap_servers=bootstrap_servers,
            key_serializer=lambda k: json.dumps(k).encode(),
            value_serializer=lambda v: json.dumps(v).encode(),
        )

    def sendmsg(
        self,
        topic: str,
        value: dict,
        key: Optional[str] = None,
    ) -> None:
        """发送数据

        Args:
            topic: kafka topic
            value: message value. Must be type bytes, or be
                serializable to bytes via configured value_serializer. If value
                is None, key is required and message acts as a 'delete'.
            key: kafka key
        """
        # Asynchronous by default
        future = (
            self.producer.send(
                topic=topic,
                value=value,
                key=key,
            )
            .add_callback(self.on_send_success)
            .add_errback(self.on_send_error)
        )

        # Block for 'synchronous' sends
        try:
            _ = future.get(timeout=10)
            # 暂不需要日志记录成功后 _ 的 partition 和 offset，故注释掉
            # Successful result returns assigned partition and offset
            # partition = record_metadata.partition
            # offset = record_metadata.offset
            # logger.info(f"save success, partition: {partition}, offset: {offset}")
        except KafkaError:
            # Decide what to do if produce request failed...
            logger.error(f"save error, topic: {topic}, value: {value}, key: {key}")

    def on_send_success(self, *args, **kwargs):
        """发送成功回调函数，暂不做任何处理或提示"""
        return

    def on_send_error(self, data, key):
        """发送失败回调函数，只日志记录"""
        logger.error(f"send error, data: {data}, key: {key}")
        return

    def close_producer(self):
        if self.producer:
            self.producer.close()


class AyuKafkaPipeline:
    kp: KafkaProducerClient

    def open_spider(self, spider):
        assert hasattr(spider, "kafka_conf"), "未配置 kafka 连接信息！"
        # 如果有多个 kafka 服务地址，用逗号分隔，会在此处拆分为列表
        _bts = spider.kafka_conf.bootstrap_servers
        bts_lst = _bts.split(",")
        self.kp = KafkaProducerClient(bootstrap_servers=bts_lst)

    def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        self.kp.sendmsg(
            topic=spider.kafka_conf.topic,
            value=item_dict,
            key=spider.kafka_conf.key,
        )
        return item

    def close_spider(self, spider):
        self.kp.close_producer()



from twisted.internet import defer, reactor

from ayugespidertools.common.mongodbpipe import TwistedAsynchronous, mongodb_pipe
from ayugespidertools.common.multiplexing import ReuseOperation
from ayugespidertools.scraper.pipelines.mongo.fantasy import AyuFtyMongoPipeline

__all__ = [
    "AyuTwistedMongoPipeline",
]


class AyuTwistedMongoPipeline(AyuFtyMongoPipeline):
    @defer.inlineCallbacks
    def process_item(self, item, spider):
        out = defer.Deferred()
        reactor.callInThread(self.db_insert, item, out)
        yield out
        defer.returnValue(item)

    def db_insert(self, item, out):
        item_dict = ReuseOperation.item_to_dict(item)
        mongodb_pipe(
            TwistedAsynchronous(),
            item_dict=item_dict,
            db=self.db,
            sys_ver_low=self.sys_ver_low,
        )
        reactor.callFromThread(out.callback, item_dict)



import string
from pathlib import Path
from shutil import ignore_patterns, move

from scrapy.commands.startproject import Command
from scrapy.exceptions import UsageError
from scrapy.utils.template import render_templatefile, string_camelcase

import ayugespidertools
from ayugespidertools.common.params import Param

# 添加需要的自定义配置文件
TEMPLATES_TO_RENDER = (
    ("pyproject.toml",),
    ("README.md",),
    ("requirements.txt",),
    ("scrapy.cfg",),
    ("${project_name}", "settings.py.tmpl"),
    ("${project_name}", "items.py.tmpl"),
    ("${project_name}", "pipelines.py.tmpl"),
    ("${project_name}", "middlewares.py.tmpl"),
    # 添加 run.py 总运行文件
    ("${project_name}", "run.py.tmpl"),
)

IGNORE = ignore_patterns("*.pyc", "__pycache__", ".svn")


class AyuCommand(Command):
    def run(self, args, opts):
        if len(args) not in (1, 2):
            raise UsageError()

        project_name = args[0]

        if len(args) == 2:
            _has_project_dir_args = True
            project_dir = Path(args[1])
        else:
            _has_project_dir_args = False
            project_dir = Path(args[0])

        if (project_dir / "scrapy.cfg").exists():
            self.exitcode = 1
            print(f"Error: scrapy.cfg already exists in {project_dir.resolve()}")
            return

        if not self._is_valid_name(project_name):
            self.exitcode = 1
            return

        self._copytree(Path(self.templates_dir), project_dir.resolve())
        move(project_dir / "module", project_dir / project_name)
        for paths in TEMPLATES_TO_RENDER:
            tplfile = Path(
                project_dir,
                *(
                    string.Template(s).substitute(project_name=project_name)
                    for s in paths
                ),
            )
            render_templatefile(
                tplfile,
                project_name=project_name,
                ProjectName=string_camelcase(project_name),
            )

        # 添加执行 shell 文件 run.sh 的生成
        if _has_project_dir_args:
            run_shell_path = f"{project_dir}/{project_name}/run.sh.tmpl"
        else:
            run_shell_path = f"{project_dir}/{project_dir}/run.sh.tmpl"
        run_shell_abspath = Path(project_dir).resolve()
        # 如果是 windows 环境的话，就不生成 shell 文件了，没啥必要
        if Param.IS_WINDOWS:
            print("Info: The run.sh file is no longer generated under windows.")
            del_file = Path(run_shell_path)
            if Path.exists(del_file):
                del_file.unlink()

        else:
            render_templatefile(
                run_shell_path,
                project_startup_dir=run_shell_abspath,
                ProjectStartupDir=string_camelcase(str(run_shell_abspath)),
                project_name=project_name,
                ProjectName=string_camelcase(project_name),
            )

        print(
            f"New Scrapy project '{project_name}', using template directory "
            f"'{self.templates_dir}', created in:"
        )
        print(f"    {project_dir.resolve()}\n")
        print("You can start your first spider with:")
        print(f"    cd {project_dir}")
        print("    scrapy genspider example example.com")
        # 添加本库的文字提示内容
        print("Or you can start your first spider with ayuge:")
        print("    ayuge genspider example example.com")

    @property
    def templates_dir(self) -> str:
        # 修改 startproject 模板文件路径为 ayugespidertools 的自定义路径
        return str(
            Path(
                Path(ayugespidertools.__path__[0], "templates"),
                "project",
            )
        )



import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Tuple, TypeVar, Union

from ayugespidertools.config import logger

__all__ = [
    "Synchronize",
    "TwistedAsynchronous",
    "deal_mysql_err",
]

if TYPE_CHECKING:
    from pymysql.connections import Connection
    from pymysql.cursors import Cursor, DictCursor
    from twisted.enterprise.adbapi import Transaction

    from ayugespidertools.common.typevars import MysqlConf

    TwistedTransactionT = TypeVar("TwistedTransactionT", bound=Transaction)
    PymysqlDictCursorT = TypeVar("PymysqlDictCursorT", bound=DictCursor)


class AbstractClass(ABC):
    """用于处理 mysql 异常的模板方法类"""

    def _create_table(
        self,
        cursor: "Cursor",
        table_name: str,
        engine: str,
        charset: str,
        collate: str,
        table_notes: str = "",
    ) -> None:
        """创建数据库表

        Args:
            cursor: mysql connect cursor
            table_name: 创建表的名称
            engine: 创建表的 engine
            charset: charset
            collate: collate
            table_notes: 创建表的注释
        """
        sql = (
            f"CREATE TABLE IF NOT EXISTS `{table_name}` (`id` int(32) NOT NULL"
            f" AUTO_INCREMENT COMMENT 'id', PRIMARY KEY (`id`)) ENGINE={engine}"
            f" DEFAULT CHARSET={charset} COLLATE={collate} COMMENT='{table_notes}';"
        )

        try:
            cursor.execute(sql)
            logger.info(f"创建数据表 {table_notes}: {table_name} 成功！")
        except Exception as e:
            logger.error(f"创建表 {table_name} 失败，err：{e}")

    def _get_column_type(
        self,
        cursor: "Cursor",
        database: str,
        table: str,
        column: str,
    ) -> Union[str, None]:
        """获取数据字段存储类型

        Args:
            cursor: mysql connect cursor
            database: 数据库名
            table: 数据表名
            column: 字段名称

        Returns:
            column_type: 字段存储类型
        """
        sql = (
            f"select COLUMN_TYPE from information_schema.columns where table_schema = '{database}'"
            f" and table_name = '{table}' and COLUMN_NAME= '{column}';"
        )
        column_type = None
        try:
            cursor.execute(sql)
            lines = cursor.fetchall()
            if isinstance(lines, list):
                # 注意，此处 AyuMysqlPipeline 返回的结构示例为：[{'COLUMN_TYPE': 'varchar(190)'}]
                column_type = lines[0]["COLUMN_TYPE"] if len(lines) == 1 else ""
            else:
                # 注意，此处 AyuTwistedMysqlPipeline 返回的结构示例为：(('varchar(10)',),)
                column_type = lines[0][0] if len(lines) == 1 else ""

        except Exception as e:
            logger.error(f"未获取到当前字段的存储类型，err: {e}")
        return column_type

    def template_method(
        self,
        err_msg: str,
        conn: "Connection[Cursor]",
        cursor: Union["Cursor", "TwistedTransactionT"],
        mysql_conf: "MysqlConf",
        table: str,
        table_notes: str,
        note_dic: dict,
    ) -> None:
        """模板方法，用于处理 mysql 存储场景的异常

        Args:
            err_msg: pipeline 存储时报错内容
            conn: mysql conn
            cursor: mysql connect cursor
            mysql_conf: spider mysql_conf
            table: 数据表
            table_notes: 数据表注释
            note_dic: 当前表字段注释
        """
        if "1054" in err_msg:
            sql, possible_err = self.deal_1054_error(
                err_msg=err_msg, table=table, note_dic=note_dic
            )
            self._exec_sql(conn=conn, cursor=cursor, sql=sql, possible_err=possible_err)

        elif "1146" in err_msg:
            self._create_table(
                cursor=cursor,
                table_name=table,
                engine=mysql_conf.engine,
                charset=mysql_conf.charset,
                collate=mysql_conf.collate,
                table_notes=table_notes,
            )

        elif "1406" in err_msg:
            sql, possible_err = self.deal_1406_error(
                err_msg=err_msg,
                cursor=cursor,
                database=mysql_conf.database,
                table=table,
                note_dic=note_dic,
            )
            self._exec_sql(conn=conn, cursor=cursor, sql=sql, possible_err=possible_err)

        elif "1265" in err_msg:
            sql, possible_err = self.deal_1265_error(
                err_msg=err_msg,
                cursor=cursor,
                database=mysql_conf.database,
                table=table,
                note_dic=note_dic,
            )
            self._exec_sql(conn=conn, cursor=cursor, sql=sql, possible_err=possible_err)

        else:
            raise Exception(f"MYSQL OTHER ERROR: {err_msg}")

    def deal_1054_error(
        self, err_msg: str, table: str, note_dic: dict
    ) -> Tuple[str, str]:
        """解决 1054, u"Unknown column 'xx' in 'field list'"

        Args:
            err_msg: 报错内容
            table: 数据表名
            note_dic: 当前表字段的注释

        Returns:
            1). sql: 用于添加字段的 sql 语句
            2). 执行此 sql 可能会报错的信息
        """
        colum_pattern = re.compile(r"Unknown column '(.*?)' in 'field list'")
        text = re.findall(colum_pattern, err_msg)
        colum = text[0]
        notes = note_dic[colum]

        sql = f"ALTER TABLE `{table}` ADD COLUMN `{colum}` VARCHAR(255) NULL DEFAULT '' COMMENT '{notes}';"
        return sql, f"添加字段 {colum} 已存在"

    def deal_1406_error(
        self,
        err_msg: str,
        cursor: "Cursor",
        database: str,
        table: str,
        note_dic: dict,
    ) -> Tuple[str, str]:
        """解决 1406, u"Data too long for 'xx' at ..."

        Args:
            err_msg: 报错内容
            cursor: mysql connect cursor
            database: 数据库名
            table: 数据表名
            note_dic: 当前表字段的注释

        Returns:
            1). sql: 修改字段类型的 sql
            2). 执行此 sql 可能会报错的信息
        """
        if "Data too long for" in err_msg:
            colum_pattern = re.compile(r"Data too long for column '(.*?)' at")
            text = re.findall(colum_pattern, err_msg)
            colum = text[0]
            notes = note_dic[colum]
            column_type = self._get_column_type(
                cursor=cursor, database=database, table=table, column=colum
            )
            change_colum_type = "LONGTEXT" if column_type == "text" else "TEXT"
            sql = (
                f"ALTER TABLE `{table}` CHANGE COLUMN `{colum}` `{colum}`"
                f' {change_colum_type} NULL DEFAULT NULL COMMENT "{notes}";'
            )
            return sql, f"更新 {colum} 字段类型为 {change_colum_type} 时失败"
        raise Exception(f"未解决 Data too long 的问题，err: {err_msg}")

    def deal_1265_error(
        self,
        err_msg: str,
        cursor: "Cursor",
        database: str,
        table: str,
        note_dic: dict,
    ) -> Tuple[str, str]:
        """解决 1265, u"Data truncated for column 'xx' at ..."

        Args:
            err_msg: 报错内容
            cursor: mysql connect cursor
            database: 数据库名
            table: 数据表名
            note_dic: 当前表字段的注释

        Returns:
            1). sql: 修改字段类型的 sql
            2). 执行此 sql 可能会报错的信息
        """
        if "Data truncated for column" in err_msg:
            colum_pattern = re.compile(r"Data truncated for column '(.*?)' at")
            text = re.findall(colum_pattern, err_msg)
            colum = text[0]
            notes = note_dic[colum]
            column_type = self._get_column_type(
                cursor=cursor, database=database, table=table, column=colum
            )
            change_colum_type = "LONGTEXT" if column_type == "text" else "TEXT"
            sql = (
                f"ALTER TABLE `{table}` CHANGE COLUMN `{colum}` `{colum}`"
                f' {change_colum_type} NULL DEFAULT NULL COMMENT "{notes}";'
            )
            return sql, f"更新 {colum} 字段类型为 {change_colum_type} 时失败"
        raise Exception(f"未解决 Data truncated 问题，err: {err_msg}")

    @abstractmethod
    def _exec_sql(self, *args, **kwargs) -> None:
        """子类要实现执行 sql 的不同方法，使得可以正常适配不同的 pipelines 场景"""
        pass


class Synchronize(AbstractClass):
    """pipeline 同步执行 sql 的场景"""

    def _exec_sql(
        self,
        conn: "Connection[Cursor]",
        cursor: "Cursor",
        sql: str,
        possible_err: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        try:
            cursor.execute(sql)
        except Exception as e:
            logger.warning(
                f"synchronize mysql exec sql err: {str(e)}\n"
                f"possible_err ->: {possible_err}"
            )


class TwistedAsynchronous(AbstractClass):
    """pipeline twisted 异步执行 sql 的场景"""

    def _exec_sql(
        self,
        cursor: "TwistedTransactionT",
        sql: str,
        possible_err: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        try:
            cursor.execute(sql)
        except Exception as e:
            logger.warning(
                f"twisted mysql exec sql err: {str(e)}\n"
                f"possible_err: {possible_err}"
            )


def deal_mysql_err(
    abstract_class: AbstractClass,
    err_msg: str,
    cursor: Union["Cursor", "TwistedTransactionT"],
    mysql_conf: "MysqlConf",
    table: str,
    table_notes: str,
    note_dic: dict,
    conn: Optional["Connection[Cursor]"] = None,
) -> None:
    abstract_class.template_method(
        err_msg,
        conn,
        cursor,
        mysql_conf,
        table,
        table_notes,
        note_dic,
    )



from typing import TYPE_CHECKING, Optional, Union

from scrapy import FormRequest

from ayugespidertools.scraper.http.request import AiohttpRequest

__all__ = [
    "AiohttpFormRequest",
]

if TYPE_CHECKING:
    from ayugespidertools.common.typevars import AiohttpRequestArgs


class AiohttpFormRequest(AiohttpRequest, FormRequest):
    """使用 aiohttp 发送 FormRequest 请求"""

    def __init__(
        self,
        url=None,
        callback=None,
        method=None,
        formdata=None,
        body=None,
        args: Optional[Union["AiohttpRequestArgs", dict]] = None,
        **kwargs
    ):
        # First init FormRequest to get url, body and method
        if formdata:
            FormRequest.__init__(self, url=url, method=method, formdata=formdata)
            url, method, body = self.url, self.method, self.body
        # Then pass all other kwargs to AiohttpRequest
        AiohttpRequest.__init__(
            self,
            url=url,
            callback=callback,
            method=method,
            body=body,
            args=args,
            **kwargs
        )



from typing import TYPE_CHECKING

from twisted.enterprise import adbapi

from ayugespidertools.common.expend import OraclePipeEnhanceMixin
from ayugespidertools.common.multiplexing import ReuseOperation

__all__ = [
    "AyuTwistedOraclePipeline",
]

if TYPE_CHECKING:
    from oracledb.connection import Connection

    from ayugespidertools.common.typevars import OracleConf, slogT


class AyuTwistedOraclePipeline(OraclePipeEnhanceMixin):
    oracle_conf: "OracleConf"
    slog: "slogT"
    conn: "Connection"
    dbpool: "adbapi.ConnectionPool"

    def open_spider(self, spider):
        assert hasattr(spider, "oracle_conf"), "未配置 Oracle 连接信息！"
        self.slog = spider.slog
        self.oracle_conf = spider.oracle_conf

        _oracle_conf = {
            "user": self.oracle_conf.user,
            "password": self.oracle_conf.password,
            "host": self.oracle_conf.host,
            "port": self.oracle_conf.port,
            "service_name": self.oracle_conf.service_name,
            "encoding": self.oracle_conf.encoding,
            "config_dir": self.oracle_conf.thick_lib_dir or None,
        }
        self.dbpool = adbapi.ConnectionPool(
            "oracledb", cp_reconnect=True, **_oracle_conf
        )
        query = self.dbpool.runInteraction(self.db_create)
        query.addErrback(self.db_create_err)

    def db_create(self, cursor):
        pass

    def db_create_err(self, failure):
        self.slog.error(f"创建数据表失败: {failure}")

    def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        query = self.dbpool.runInteraction(self.db_insert, item_dict)
        query.addErrback(self.handle_error, item)
        return item

    def db_insert(self, cursor, item):
        alter_item = ReuseOperation.reshape_item(item)
        if not (new_item := alter_item.new_item):
            return

        sql = self._get_sql_by_item(table=alter_item.table.name, item=new_item)
        cursor.execute(sql, tuple(new_item.values()))
        return item

    def handle_error(self, failure, item):
        self.slog.error(f"插入数据失败:{failure}, item: {item}")



from typing import TYPE_CHECKING

import pymysql
from dbutils.pooled_db import PooledDB

from ayugespidertools.scraper.pipelines.mysql import AyuMysqlPipeline

__all__ = [
    "AyuTurboMysqlPipeline",
]

if TYPE_CHECKING:
    from pymysql.connections import Connection
    from pymysql.cursors import Cursor

    from ayugespidertools.common.typevars import MysqlConf, slogT


class AyuTurboMysqlPipeline(AyuMysqlPipeline):
    mysql_conf: "MysqlConf"
    conn: "Connection[Cursor]"
    slog: "slogT"
    cursor: "Cursor"
    pool_db_conf: dict

    def __init__(self, pool_db_conf):
        self.pool_db_conf = pool_db_conf

    @classmethod
    def from_crawler(cls, crawler):
        pool_db_conf = crawler.settings.get("POOL_DB_CONFIG", None)
        return cls(
            pool_db_conf=pool_db_conf,
        )

    def open_spider(self, spider):
        assert hasattr(spider, "mysql_conf"), "未配置 Mysql 连接信息！"
        self.slog = spider.slog
        if not self.pool_db_conf:
            spider.slog.warning("未配置 POOL_DB_CONFIG 参数，将使用其默认参数")
            self.pool_db_conf = {
                "maxconnections": 5,
                # 连接池中空闲连接的最大数量。默认 0，即无最大数量限制
                "maxcached": 0,
                # 连接的最大使用次数。默认 0，即无使用次数限制
                "maxusage": 0,
                # 连接数达到最大时，新连接是否可阻塞。默认False，即达到最大连接数时，再取新连接将会报错
                "blocking": True,
            }
        self.mysql_conf = spider.mysql_conf

        # 判断目标数据库是否连接正常。若连接目标数据库错误时，创建缺失的目标数据库。这个并不需要此连接对象，直接关闭即可
        self._connect(spider.mysql_conf).close()

        # 添加 PooledDB 的配置
        self.conn = PooledDB(
            creator=pymysql,
            user=self.mysql_conf.user,
            password=self.mysql_conf.password,
            host=self.mysql_conf.host,
            port=self.mysql_conf.port,
            database=self.mysql_conf.database,
            charset=self.mysql_conf.charset,
            **self.pool_db_conf
        ).connection()
        self.cursor = self.conn.cursor()



import json
from typing import TYPE_CHECKING, Union

import pika

from ayugespidertools.common.multiplexing import ReuseOperation
from ayugespidertools.items import AyuItem

__all__ = [
    "AyuMQPipeline",
]

if TYPE_CHECKING:
    from pika.adapters.blocking_connection import BlockingChannel, BlockingConnection

    from ayugespidertools.common.typevars import MQConf


class AyuMQPipeline:
    channel: "BlockingChannel"
    conn: "BlockingConnection"

    @staticmethod
    def _dict_to_bytes(item: Union[AyuItem, dict]) -> bytes:
        item_dict = ReuseOperation.item_to_dict(item)
        item_json_str = json.dumps(item_dict)
        return bytes(item_json_str, encoding="utf-8")

    def open_spider(self, spider):
        assert hasattr(spider, "rabbitmq_conf"), "未配置 RabbitMQ 连接信息！"
        _mq_conf: "MQConf" = spider.rabbitmq_conf
        mq_conn_param = pika.URLParameters(
            f"amqp://{_mq_conf.username}:{_mq_conf.password}"
            f"@{_mq_conf.host}:{_mq_conf.port}/{_mq_conf.virtualhost}"
            f"?heartbeat={_mq_conf.heartbeat}&socket_timeout={_mq_conf.socket_timeout}"
        )
        self.conn = pika.BlockingConnection(parameters=mq_conn_param)
        self.channel = self.conn.channel()
        self.channel.queue_declare(
            queue=_mq_conf.queue,
            durable=_mq_conf.durable,
            exclusive=_mq_conf.exclusive,
            auto_delete=_mq_conf.auto_delete,
        )
        self.channel.confirm_delivery()

    def close_spider(self, spider):
        self.conn.close()

    def process_item(self, item, spider):
        self.channel.basic_publish(
            exchange=spider.rabbitmq_conf.exchange,
            routing_key=spider.rabbitmq_conf.routing_key,
            body=self._dict_to_bytes(item),
            properties=pika.BasicProperties(
                content_type=spider.rabbitmq_conf.content_type,
                delivery_mode=spider.rabbitmq_conf.delivery_mode,
            ),
            mandatory=True,
        )
        return item



import datetime
from typing import TYPE_CHECKING, Tuple

from retrying import retry

from ayugespidertools.common.expend import MysqlPipeEnhanceMixin
from ayugespidertools.common.params import Param

__all__ = [
    "AyuStatisticsMysqlPipeline",
]

if TYPE_CHECKING:
    from pymysql.connections import Connection
    from pymysql.cursors import Cursor

    from ayugespidertools.common.typevars import MysqlConf, slogT


class AyuStatisticsMysqlPipeline(MysqlPipeEnhanceMixin):
    mysql_conf: "MysqlConf"
    conn: "Connection[Cursor]"
    slog: "slogT"
    cursor: "Cursor"
    crawl_time: "datetime.date"

    def open_spider(self, spider):
        self.crawl_time = datetime.date.today()
        self.slog = spider.slog
        self.mysql_conf = spider.mysql_conf
        self.conn = self._connect(self.mysql_conf)
        self.cursor = self.conn.cursor()

    def table_collection_statistics(
        self, spider_name: str, database: str, crawl_time: datetime.date
    ):
        """统计数据库入库数据，获取当前数据库中所有包含 crawl_time 字段的数据表的简要信息

        Args:
            spider_name: 爬虫脚本名称
            database: 数据库，保存程序采集记录保存的数据库
            crawl_time: 采集时间，程序运行时间
        """
        sql = f"""
        select concat(
        'select "', TABLE_NAME, '", count(id) as num , crawl_time from ', TABLE_SCHEMA, '.', TABLE_NAME,
        ' where crawl_time = "{crawl_time}"') from information_schema.tables
        where TABLE_SCHEMA='{database}' and TABLE_NAME in
        (SELECT TABLE_NAME FROM information_schema.columns WHERE COLUMN_NAME='crawl_time');
        """
        self.cursor.execute(sql)
        results = self.cursor.fetchall()
        if sql_list := [row[0] for row in results]:
            sql_all = " union all ".join(sql_list)
            self.cursor.execute(sql_all)
            results = self.cursor.fetchall()

            for row in results:
                table_statistics = {
                    "spider_name": spider_name,
                    "database": database,
                    "table_name": row[0],
                    "number": row[1],
                    "crawl_time": str((row[2] or crawl_time)),
                }
                self.insert_table_statistics(table_statistics)

    def insert_table_statistics(
        self, data: dict, table: str = "table_collection_statistics"
    ):
        """插入统计数据到表中

        Args:
            data: 需要统计的入库信息
            table: 存储表的名称
        """
        create_table_sql = f"""
        CREATE TABLE IF NOT EXISTS `{table}` (
            `id` int(11) NOT NULL AUTO_INCREMENT,
            `database` varchar(255) NOT NULL DEFAULT '-' COMMENT '采集程序和记录信息存储的数据库名',
            `spider_name` varchar(255) NOT NULL DEFAULT '-' COMMENT '脚本名称',
            `crawl_time` datetime NOT NULL COMMENT '程序运行/数据采集时间',
            `table_name` varchar(255) NOT NULL COMMENT '此项目所在库（一般某个项目放在单独的数据库中）的当前表名',
            `number` varchar(255) NOT NULL COMMENT '当前表的当前 crawl_time 的采集个数',
            PRIMARY KEY (`id`) USING BTREE
        ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC COMMENT='项目对应库中各表采集统计表';
        """
        self.cursor.execute(create_table_sql)

        sql, args = self._get_sql_by_item(
            table=table,
            item=data,
            odku_enable=self.mysql_conf.odku_enable,
        )
        self._log_record(sql=sql, data=args)

    @retry(
        stop_max_attempt_number=Param.retry_num,
        wait_random_min=Param.retry_time_min,
        wait_random_max=Param.retry_time_max,
    )
    def insert_script_statistics(
        self, data: dict, table: str = "script_collection_statistics"
    ):
        """存储运行脚本的统计信息

        Args:
            data: 需要插入的 log 信息
            table: 存储表的名称
        """
        self.cursor.execute(
            f"""
                CREATE TABLE IF NOT EXISTS `{table}` (
                    `id` int(11) NOT NULL AUTO_INCREMENT,
                    `database` varchar(255) NOT NULL DEFAULT '-' COMMENT '采集程序和记录信息存储的数据库名',
                    `spider_name` varchar(255) NOT NULL DEFAULT '-' COMMENT '脚本名称',
                    `uid` varchar(255) NOT NULL DEFAULT '-' COMMENT 'uid',
                    `request_counts` varchar(255) NOT NULL DEFAULT '-' COMMENT '请求次数统计',
                    `received_count` varchar(255) NOT NULL DEFAULT '-' COMMENT '接收次数统计',
                    `item_counts` varchar(255) NOT NULL DEFAULT '-' COMMENT '采集数据量',
                    `info_count` varchar(255) NOT NULL DEFAULT '-' COMMENT 'info 数据统计',
                    `warning_count` varchar(255) NOT NULL DEFAULT '-' COMMENT '警告数据统计',
                    `error_count` varchar(255) NOT NULL DEFAULT '-' COMMENT '错误数据统计',
                    `start_time` datetime NOT NULL COMMENT '开始时间',
                    `finish_time` datetime NOT NULL COMMENT '结束时间',
                    `spend_minutes` varchar(255) NOT NULL DEFAULT '-' COMMENT '花费时间',
                    `crawl_time` datetime NOT NULL COMMENT '程序运行/数据采集时间',
                    `log_count_ERROR` varchar(255) DEFAULT NULL COMMENT '错误原因',
                    PRIMARY KEY (`id`) USING BTREE
                    ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC COMMENT='项目运行脚本统计信息表';
                """
        )

        sql, args = self._get_sql_by_item(
            table=table,
            item=data,
            odku_enable=self.mysql_conf.odku_enable,
        )
        self._log_record(sql=sql, data=args)

    def _log_record(self, sql: str, data: Tuple) -> None:
        """执行日志记录的 sql 语句

        Args:
            sql: sql 语句
            data: sql 语句中的参数
        """
        try:
            self.cursor.execute(sql, data)
            self.conn.commit()
        except Exception as e:
            self.conn.rollback()
            self.slog.warning(f"日志记录存储错误: {e}")

    def close_spider(self, spider):
        log_info = self._get_log_by_spider(spider=spider, crawl_time=self.crawl_time)

        # 运行脚本统计信息
        self.insert_script_statistics(log_info)
        self.table_collection_statistics(
            spider_name=spider.name,
            database=spider.mysql_conf.database,
            crawl_time=self.crawl_time,
        )

        if self.conn:
            self.conn.close()

    def process_item(self, item, spider):
        return item



# Define your item pipelines here
#
# Don"t forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html
from ayugespidertools.scraper.pipelines.download.file import FilesDownloadPipeline
from ayugespidertools.scraper.pipelines.es.asynced import AyuAsyncESPipeline
from ayugespidertools.scraper.pipelines.es.fantasy import AyuFtyESPipeline
from ayugespidertools.scraper.pipelines.mongo.asynced import AyuAsyncMongoPipeline
from ayugespidertools.scraper.pipelines.mongo.fantasy import AyuFtyMongoPipeline
from ayugespidertools.scraper.pipelines.mongo.twisted import AyuTwistedMongoPipeline
from ayugespidertools.scraper.pipelines.msgproducer.mqpub import AyuMQPipeline
from ayugespidertools.scraper.pipelines.mysql import AyuMysqlPipeline
from ayugespidertools.scraper.pipelines.mysql.asynced import AyuAsyncMysqlPipeline
from ayugespidertools.scraper.pipelines.mysql.fantasy import AyuFtyMysqlPipeline
from ayugespidertools.scraper.pipelines.mysql.stats import AyuStatisticsMysqlPipeline
from ayugespidertools.scraper.pipelines.mysql.turbo import AyuTurboMysqlPipeline
from ayugespidertools.scraper.pipelines.mysql.twisted import AyuTwistedMysqlPipeline
from ayugespidertools.scraper.pipelines.oracle.fantasy import AyuFtyOraclePipeline
from ayugespidertools.scraper.pipelines.oracle.twisted import AyuTwistedOraclePipeline
from ayugespidertools.scraper.pipelines.oss.ali import AyuAsyncOssPipeline
from ayugespidertools.scraper.pipelines.postgres.asynced import AyuAsyncPostgresPipeline
from ayugespidertools.scraper.pipelines.postgres.fantasy import AyuFtyPostgresPipeline
from ayugespidertools.scraper.pipelines.postgres.twisted import (
    AyuTwistedPostgresPipeline,
)

__all__ = [
    "FilesDownloadPipeline",
    "AyuFtyESPipeline",
    "AyuAsyncESPipeline",
    "AyuAsyncMongoPipeline",
    "AyuFtyMongoPipeline",
    "AyuTwistedMongoPipeline",
    "AyuMQPipeline",
    "AyuMysqlPipeline",
    "AyuAsyncMysqlPipeline",
    "AyuFtyMysqlPipeline",
    "AyuStatisticsMysqlPipeline",
    "AyuTurboMysqlPipeline",
    "AyuTwistedMysqlPipeline",
    "AyuFtyOraclePipeline",
    "AyuTwistedOraclePipeline",
    "AyuAsyncOssPipeline",
    "AyuAsyncPostgresPipeline",
    "AyuFtyPostgresPipeline",
    "AyuTwistedPostgresPipeline",
]



from scrapy.commands.version import Command

from ayugespidertools import __version__


class AyuCommand(Command):
    def short_desc(self):
        return "Print AyugeSpiderTools version"

    def run(self, args, opts):
        print(f"AyugeSpiderTools {__version__}")



from typing import TYPE_CHECKING

from ayugespidertools.common.expend import PostgreSQLPipeEnhanceMixin
from ayugespidertools.common.multiplexing import ReuseOperation
from ayugespidertools.common.postgreserrhandle import Synchronize, deal_postgres_err

__all__ = ["AyuPostgresPipeline"]

if TYPE_CHECKING:
    from psycopg.connection import Connection
    from psycopg.cursor import Cursor

    from ayugespidertools.common.typevars import AlterItem, slogT


class AyuPostgresPipeline(PostgreSQLPipeEnhanceMixin):
    conn: "Connection"
    slog: "slogT"
    cursor: "Cursor"

    def open_spider(self, spider):
        assert hasattr(spider, "postgres_conf"), "未配置 PostgreSQL 连接信息！"
        self.slog = spider.slog
        self.conn = self._connect(spider.postgres_conf)
        self.cursor = self.conn.cursor()

    def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        alter_item = ReuseOperation.reshape_item(item_dict)
        self.insert_item(alter_item)
        return item

    def insert_item(self, alter_item: "AlterItem"):
        if not (new_item := alter_item.new_item):
            return

        _table_name = alter_item.table.name
        _table_notes = alter_item.table.notes
        note_dic = alter_item.notes_dic
        sql = self._get_sql_by_item(table=_table_name, item=new_item)

        try:
            self.cursor.execute(sql, tuple(new_item.values()))
            self.conn.commit()

        except Exception as e:
            self.slog.warning(
                f"Pipe Warn: {e} & Table: {_table_name} & Item: {new_item}"
            )
            self.conn.rollback()
            deal_postgres_err(
                Synchronize(),
                err_msg=str(e),
                conn=self.conn,
                cursor=self.cursor,
                table=_table_name,
                table_notes=_table_notes,
                note_dic=note_dic,
            )
            return self.insert_item(alter_item)

    def close_spider(self, spider):
        self.cursor.close()
        self.conn.close()



import asyncio
from typing import TYPE_CHECKING, Optional, Tuple, TypeVar

import aiohttp
import scrapy
from itemadapter import ItemAdapter
from scrapy.http import HtmlResponse
from scrapy.utils.python import global_object_name

from ayugespidertools.common.multiplexing import ReuseOperation
from ayugespidertools.common.params import Param
from ayugespidertools.common.typevars import AiohttpConf, AiohttpRequestArgs
from ayugespidertools.common.utils import ToolsForAyu
from ayugespidertools.config import logger

__all__ = [
    "AiohttpDownloaderMiddleware",
]

if TYPE_CHECKING:
    from aiohttp.connector import BaseConnector

ItemAdapterT = TypeVar("ItemAdapterT", bound=ItemAdapter)


class AiohttpDownloaderMiddleware:
    """Downloader middleware handling the requests with aiohttp"""

    def __init__(self):
        self.aiohttp_args = None

    def _retry(
        self,
        request: scrapy.Request,
        reason: int,
        spider: scrapy.Spider,
    ) -> Optional[scrapy.Request]:
        """重试请求

        Args:
            request: scrapy request
            reason: reason
            spider: scrapy spider

        Returns:
            Optional[scrapy.Request]: 重试的 request 对象
        """
        retries = request.meta.get("retry_times", 0) + 1
        stats = spider.crawler.stats
        if retries <= self.retry_times:
            return self._retry_with_limit(request, retries, reason, stats)

        stats.inc_value("retry/max_reached")
        logger.error(f"Gave up retrying {request} (failed {retries} times): {reason}")
        return None

    def _retry_with_limit(self, request, retries, reason, stats):
        logger.debug(f"Retrying {request} (failed {retries} times): {reason}")
        retry_req = request.copy()
        retry_req.meta["retry_times"] = retries
        retry_req.dont_filter = True
        # 优先级逐级降低，以防堆积
        retry_req.priority = request.priority + self.priority_adjust

        if isinstance(reason, Exception):
            reason = global_object_name(reason.__class__)

        stats.inc_value("retry/count")
        stats.inc_value(f"retry/reason_count/{reason}")
        return retry_req

    def _get_args(self, key: str):
        """根据优先级依次获取不为 None 的请求参数"""
        data_lst = [
            self.aiohttp_args.get(key),
            getattr(self, key),
        ]
        return ToolsForAyu.first_not_none(data_lst=data_lst)

    @classmethod
    def from_crawler(cls, crawler):
        settings = crawler.settings
        # 自定义 aiohttp 全局配置信息，优先级小于 aiohttp_meta 中的配置
        if local_aiohttp_conf := settings.get("AIOHTTP_CONFIG", {}):
            # 这里的配置信息如果在 aiohttp_meta 中重复设置，则会更新当前请求的参数
            _aiohttp_conf = AiohttpConf(
                timeout=settings.get("DOWNLOAD_TIMEOUT"),
                sleep=local_aiohttp_conf.get("sleep"),
                proxy=local_aiohttp_conf.get("proxy"),
                proxy_auth=local_aiohttp_conf.get("proxy_auth"),
                proxy_headers=local_aiohttp_conf.get("proxy_headers"),
                retry_times=local_aiohttp_conf.get(
                    "retry_times", Param.aiohttp_retry_times_default
                ),
                limit=local_aiohttp_conf.get("limit", 100),
                ssl=local_aiohttp_conf.get("ssl", True),
                verify_ssl=local_aiohttp_conf.get("verify_ssl", True),
                limit_per_host=local_aiohttp_conf.get("limit_per_host", 0),
                allow_redirects=local_aiohttp_conf.get("allow_redirects", True),
            )

            # 初始化所需要的参数信息，用于构建 aiohttp 请求信息
            # cls.retry_http_codes = set(int(x) for x in settings.getlist('RETRY_HTTP_CODES'))
            cls.timeout = _aiohttp_conf.timeout
            cls.proxy = _aiohttp_conf.proxy
            cls.proxy_auth = _aiohttp_conf.proxy_auth
            cls.proxy_headers = _aiohttp_conf.proxy_headers
            cls.sleep = _aiohttp_conf.sleep
            cls.retry_times = _aiohttp_conf.retry_times
            cls.limit = _aiohttp_conf.limit
            cls.ssl = _aiohttp_conf.ssl
            cls.verify_ssl = _aiohttp_conf.verify_ssl
            cls.limit_per_host = _aiohttp_conf.limit_per_host
            cls.allow_redirects = _aiohttp_conf.allow_redirects
            cls.priority_adjust = settings.getint("RETRY_PRIORITY_ADJUST")
        return cls()

    async def _request_by_aiohttp(
        self,
        aio_request_args: ItemAdapterT,
        timeout: Optional[aiohttp.ClientTimeout] = None,
        connector: Optional["BaseConnector"] = None,
    ) -> Tuple[int, str]:
        """使用 aiohttp 来请求

        Args:
            aio_request_args: 普通的 aiohttp 请求参数
            timeout: aiohttp.ClientSession 的 timeout 参数
            connector: aiohttp connector 参数

        Returns:
            1). status_code: 请求状态码
            2). response_text: 请求返回的文本内容
        """
        try:
            async with aiohttp.ClientSession(
                connector=connector, timeout=timeout
            ) as session:
                async with session.request(**aio_request_args) as response:
                    status_code = response.status
                    response_text = await response.text()
                    return status_code, response_text
        except Exception:
            return 504, ""

    async def process_request(self, request, spider):
        aiohttp_options = request.meta.get("aiohttp")
        self.aiohttp_args = aiohttp_options.setdefault("args", {})

        # 根据 LOCAL_AIOHTTP_CONFIG 中设置 aiohttp 请求参数
        # 这些参数全局生效，不会在 meta 中更新
        _connector = aiohttp.TCPConnector(
            ssl=self.ssl,
            limit=self.limit,
            verify_ssl=self.verify_ssl,
            limit_per_host=self.limit_per_host,
        )

        # 设置 url
        _url = self.aiohttp_args.get("url") or request.url
        aiohttp_req_args = AiohttpRequestArgs(
            url=_url,
        )

        # 设置请求方式
        if _method := self.aiohttp_args.get("method"):
            aiohttp_req_args.method = _method
        elif _method := str(request.method).upper():
            aiohttp_req_args.method = _method
        if _method not in {"GET", "POST"}:
            logger.error(f"出现未知请求方式 {_method}，请及时查看，默认 GET")

        # 设置请求头信息
        if _headers_args := self.aiohttp_args.get("headers"):
            aiohttp_req_args.headers = _headers_args
        elif _headers_args := ToolsForAyu.get_dict_form_scrapy_req_headers(
            scrapy_headers=request.headers
        ):
            aiohttp_req_args.headers = _headers_args

        # 设置请求 body 参数，GET 情况下和 POST 情况下的请求参数处理
        if _req_data := self.aiohttp_args.get("data"):
            aiohttp_req_args.data = _req_data
        elif req_data_str := str(request.body, encoding="utf-8"):
            # 如果是 json 字典格式的数据时，则是 scrapy body 传来的 json dumps 参数
            if ReuseOperation.judge_str_is_json(judge_str=req_data_str):
                aiohttp_req_args.data = req_data_str
            # 否则就是传来的字典格式
            else:
                req_data_dict = ReuseOperation.get_req_dict_from_scrapy(
                    req_body_data_str=req_data_str
                )
                aiohttp_req_args.data = req_data_dict

        # 设置 proxy
        if _proxy := self._get_args("proxy"):
            aiohttp_req_args.proxy = _proxy
        if _proxy_auth := self._get_args("proxy_auth"):
            aiohttp_req_args.proxy_auth = _proxy_auth
        if _proxy_headers := self._get_args("proxy_headers"):
            aiohttp_req_args.proxy_headers = _proxy_headers

        # 设置 allow_redirects
        _allow_redirects = self._get_args("allow_redirects")
        if _allow_redirects is not None:
            aiohttp_req_args.allow_redirects = _allow_redirects

        # 设置 cookies，优先从 AiohttpRequest 中的 cookies 参数中取值，没有时再从 headers 中取值
        _ck_args = self.aiohttp_args.get("cookies")
        aiohttp_cookie_dict = _ck_args if _ck_args is not None else request.cookies
        if all([not aiohttp_cookie_dict, request.headers.get("Cookie", None)]):
            headers_cookie_str = str(request.headers.get("Cookie"), encoding="utf-8")
            aiohttp_cookie_dict = ReuseOperation.get_ck_dict_from_headers(
                headers_ck_str=headers_cookie_str
            )
        aiohttp_req_args.cookies = aiohttp_cookie_dict

        # 请求超时设置
        _timeout_obj = None
        if _timeout := self.aiohttp_args.get("timeout"):
            aiohttp_req_args.timeout = _timeout
        elif self.timeout is not None:
            _timeout_obj = aiohttp.ClientTimeout(total=self.timeout)

        aio_request_args = ItemAdapter(aiohttp_req_args)
        status_code, html_content = await self._request_by_aiohttp(
            timeout=_timeout_obj,
            aio_request_args=aio_request_args,
            connector=_connector,
        )

        # 请求间隔设置
        _sleep = self._get_args("sleep")
        await asyncio.sleep(_sleep)

        if all([status_code == 504, not html_content]):
            spider.slog.error(f"url: {_url} 返回内容为空，请求超时！")
            self._retry(request=request, reason=504, spider=spider)

        body = str.encode(html_content)
        return HtmlResponse(
            url=_url,
            status=status_code,
            headers=aiohttp_req_args.headers,
            body=body,
            encoding="utf-8",
            request=request,
        )



from ayugespidertools.items import AyuItem
from ayugespidertools.scraper.http.request import AiohttpRequest
from ayugespidertools.scraper.http.request.form import AiohttpFormRequest
from ayugespidertools.scraper.spiders import AyuSpider
from ayugespidertools.scraper.spiders.crawl import AyuCrawlSpider

__all__ = [
    "AiohttpRequest",
    "AiohttpFormRequest",
    "AyuItem",
    "AyuSpider",
    "AyuCrawlSpider",
    "__version__",
]

__version__ = "3.9.6"



from ayugespidertools.scraper.http.request import AiohttpRequest
from ayugespidertools.scraper.http.request.form import AiohttpFormRequest

__all__ = [
    "AiohttpRequest",
    "AiohttpFormRequest",
]



from typing import TYPE_CHECKING

from ayugespidertools.common.expend import OraclePipeEnhanceMixin
from ayugespidertools.common.multiplexing import ReuseOperation

__all__ = ["AyuOraclePipeline"]

if TYPE_CHECKING:
    from oracledb.connection import Connection
    from oracledb.cursor import Cursor

    from ayugespidertools.common.typevars import AlterItem


class AyuOraclePipeline(OraclePipeEnhanceMixin):
    conn: "Connection"
    cursor: "Cursor"

    def open_spider(self, spider):
        assert hasattr(spider, "oracle_conf"), "未配置 Oracle 连接信息！"
        self.conn = self._connect(spider.oracle_conf)
        self.cursor = self.conn.cursor()

    def process_item(self, item, spider):
        item_dict = ReuseOperation.item_to_dict(item)
        alter_item = ReuseOperation.reshape_item(item_dict)
        self.insert_item(alter_item)
        return item

    def insert_item(self, alter_item: "AlterItem"):
        """通用插入数据

        Args:
            alter_item: 经过转变后的 item
        """
        if not (new_item := alter_item.new_item):
            return

        sql = self._get_sql_by_item(table=alter_item.table.name, item=new_item)
        # no err handing
        self.cursor.execute(sql, tuple(new_item.values()))
        self.conn.commit()

    def close_spider(self, spider):
        self.conn.close()



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Tuple, TypeVar

from ayugespidertools.common.multiplexing import ReuseOperation

__all__ = [
    "Synchronize",
    "TwistedAsynchronous",
    "AsyncioAsynchronous",
    "mongodb_pipe",
]

if TYPE_CHECKING:
    from motor.core import AgnosticDatabase
    from pymongo.database import Database

    PymongoDataBaseT = TypeVar("PymongoDataBaseT", Database, AgnosticDatabase)


class AbstractClass(ABC):
    """用于处理 mongodb pipeline 存储的模板方法类"""

    def _get_insert_data(self, item_dict: dict) -> Tuple[dict, str]:
        """获取要插入的数据，将 item 中的存储数据提取出来

        Args:
            item_dict: item 的 dict 类型

        Returns:
            insert_data: 返回 dict 格式的存储数据
        """
        insert_data = ReuseOperation.get_items_except_keys(
            dict_conf=item_dict,
            keys=["_table", "_mongo_update_rule"],
        )
        table_name = item_dict["_table"]
        judge_item = next(iter(insert_data.values()))
        # 是 namedtuple 类型
        if ReuseOperation.is_namedtuple_instance(judge_item):
            insert_data = {v: insert_data[v].key_value for v in insert_data.keys()}
            table_name = item_dict["_table"].key_value
        # 是普通的 dict 格式，则直接为 insert_data
        return insert_data, table_name

    def process_item_template(
        self,
        item_dict: dict,
        db: "PymongoDataBaseT",
        sys_ver_low: Optional[bool] = None,
    ):
        """模板方法，用于处理 mongodb pipeline 存储的模板方法类

        Args:
            item_dict: item 的 dict 类型
            db: mongodb 数据库连接
            sys_ver_low: 是否是 py3.11 以下
        """
        insert_data, table_name = self._get_insert_data(item_dict)
        self._data_storage_logic(
            db=db,
            item_dict=item_dict,
            collection_name=table_name,
            insert_data=insert_data,
            sys_ver_low=sys_ver_low,
        )

    def _default_storage(
        self,
        db: "PymongoDataBaseT",
        item_dict: dict,
        collection_name: str,
        insert_data: dict,
        sys_ver_low: Optional[bool] = None,
    ):
        if sys_ver_low:
            # 如果没有查重字段时，就直接插入数据（不去重）
            if not item_dict.get("_mongo_update_rule"):
                db[collection_name].insert(insert_data)
            else:
                db[collection_name].update(
                    item_dict["_mongo_update_rule"], {"$set": insert_data}, True
                )
        else:
            if not item_dict.get("_mongo_update_rule"):
                db[collection_name].insert_one(insert_data)
            else:
                db[collection_name].update_many(
                    item_dict["_mongo_update_rule"], {"$set": insert_data}, True
                )

    @abstractmethod
    def _data_storage_logic(
        self,
        db: "PymongoDataBaseT",
        item_dict: dict,
        collection_name: str,
        insert_data: dict,
        sys_ver_low: Optional[bool] = None,
        *args,
        **kwargs,
    ) -> None:
        """数据存储逻辑，需要子类实现

        Args:
            db: mongodb 数据库连接
            item_dict: item 的 dict 类型
            collection_name: 集合名称
            insert_data: 要插入的数据
            *args: 可变参数
            **kwargs:关键字参数
        """
        pass


class Synchronize(AbstractClass):
    """pipeline 同步执行 mongodb 存储的场景"""

    def _data_storage_logic(
        self,
        db: "PymongoDataBaseT",
        item_dict: dict,
        collection_name: str,
        insert_data: dict,
        sys_ver_low: Optional[bool] = None,
        *args,
        **kwargs,
    ) -> None:
        self._default_storage(db, item_dict, collection_name, insert_data, sys_ver_low)


class TwistedAsynchronous(AbstractClass):
    """pipeline twisted 异步执行 mongodb 存储的场景"""

    def _data_storage_logic(
        self,
        db: "PymongoDataBaseT",
        item_dict: dict,
        collection_name: str,
        insert_data: dict,
        sys_ver_low: Optional[bool] = None,
        *args,
        **kwargs,
    ) -> None:
        self._default_storage(db, item_dict, collection_name, insert_data, sys_ver_low)


class AsyncioAsynchronous(AbstractClass):
    """pipeline asyncio 异步执行 mongodb 存储的场景 - 使用 motor 实现"""

    async def _data_storage_logic(  # type: ignore[override]
        self,
        db: "PymongoDataBaseT",
        item_dict: dict,
        collection_name: str,
        insert_data: dict,
        sys_ver_low: Optional[bool] = None,
        *args,
        **kwargs,
    ):  # @override to fix mypy [override] error.
        if not item_dict.get("_mongo_update_rule"):
            await db[collection_name].insert_one(insert_data)
        else:
            await db[collection_name].update_many(
                item_dict["_mongo_update_rule"], {"$set": insert_data}, True
            )

    async def process_item_template(
        self,
        item_dict: dict,
        db: "PymongoDataBaseT",
        sys_ver_low: Optional[bool] = None,
    ):
        insert_data, table_name = self._get_insert_data(item_dict)
        await self._data_storage_logic(
            db=db,
            item_dict=item_dict,
            collection_name=table_name,
            insert_data=insert_data,
        )


def mongodb_pipe(
    abstract_class: AbstractClass,
    item_dict: dict,
    db: "PymongoDataBaseT",
    sys_ver_low: bool = True,
) -> None:
    """mongodb pipeline 存储的通用调用方法"""
    abstract_class.process_item_template(item_dict, db, sys_ver_low)
